#!/usr/bin/env python2
# -*- coding: utf-8 -*-
# add copy number, each branch has a copy number variation
# assume the genome is diploid, and not go aneuploid
"""
@authors: Xian Fan Mallory
Contacting email: fan@cs.fsu.edu
"""
# generate tree, along with copy number variation happening to the branches. write the final fasta for each leaf to a file.
# this file is a wrapper from Beta_Splitting_Model.py as of 09052018

# -*- coding: utf-8 -*- 
from anytree import Node, RenderTree
import numpy as np
import graphviz as gv
import sys
import re
from anytree.dotexport import RenderTreeGraph
from CN import CN
from Gen_Ref_Fa import gen_ref, init_ref, write_ref, read_ref

nuc_array = ['A', 'B', 'C', 'D']
# for test purpose
random = 1
cn_list1 = [0, 309, 1282, 381, 176, 1242, 13, 29, 631]
cn_list2 = [744, 319, 1292, 389, 186, 1252, 23, 39, 641]
if_del_list = [0, 1, 0, 0, 1, 1, 1, 1, 1]
allele_list = [0, 1, 0, 1, 1, 0, 1, 0, 1]
amp_num_list = [1, 0, 3, 3, 0, 0, 0, 0, 0]
#cn_list1 = [1139, 1358, 821]
#cn_list2 = [1149, 1368, 831]
#if_del_list = [1, 0, 0]
#allele_list = [0, 1, 1]
#amp_num_list = [0, 3, 1]
#cn_list1 = [203, 211, 359, 72, 638]
#cn_list2 = [213, 219, 369, 82, 648]
#if_del_list = [0, 0, 1, 0, 0]
#allele_list = [1, 1, 0, 1, 1]
#amp_num_list = [2, 1, 0, 1, 1]
#cn_list1 = [443, 712, 731]
#cn_list2 = [453, 722, 741]
#if_del_list = [0, 0, 0]
#allele_list = [1, 0, 0]
#amp_num_list = [1, 1, 2]
#cn_list1 = [398, 432, 392, 396, 641, 659]
#cn_list2 = [408, 442, 402, 406, 649, 669]
#if_del_list = [1, 1, 0, 1, 0, 1]
#allele_list = [1, 0, 1, 1, 0, 1]
#amp_num_list = [0, 0, 1, 0, 2, 0]
#cn_list1 = [10,35,8]
#cn_list2 = [20,45,40]
#if_del_list = [1,1]


# definition of one SNV
class MySNV():
    def __init__(self, ale, chr, pos, ori_nuc, new_nuc):
        #MySNV.__init__(self, ale, chr, pos, ori_nuc, new_nuc)
        self.ale = ale
        self.chr = "NA"
        self.pos = pos
        self.ori_nuc = ori_nuc
        self.new_nuc = new_nuc

class corres_coord():
    def __init__(self, r1, r2, g1, g2):
        #corres_coord.__init__(self, r1, r2, g1, g2)
        self.ref = [r1, r2]
        self.gen = [g1, g2]

class MyNode(Node):
    def __init__(self, name, parent=None):
        Node.__init__(self, name, parent)
        self.id = 0
        self.name = name
        self.parent=parent
        self.tuple=[]
        self.is_dead=False
        self.edge_length = 0
        # alelle length for each chromosome, root has the same as reference
        self.cn=[]
        self.chrlen=[]
        self.ref=[]
        self.snvs=[]
        self.corres=[]
        self.cn_summary={}
        self.cn_detail=[]
        self.parentID = -1
    def getTuple(self):
        return self.tuple
    def setDead(self):
        self.is_dead=True
    def getID(self):
        return self.id

def get_range(chr_len, min_cn_size, exp_theta, CN_LIST_ID):
    # get pos1 and pos2 for the copy number. 
    # skip the first and last "skip" nucleotides (CN does not happen in that range) 
    if random == 0:
        # no randomness
        return cn_list1[CN_LIST_ID], cn_list2[CN_LIST_ID]
        
    else:
        skip = 0
        #skip = 5000000 
        # not try any more if cannot find a good one, just use the end
        trials = 0
        max_n = 10
        cn_size = np.random.exponential(exp_theta) + min_cn_size
        p1 = np.random.uniform(skip, chr_len - skip)
        # make sure p2 not out of boundary
        while chr_len - p1 - skip - cn_size < 0 and trials < max_n:
            p1 = np.random.uniform(skip, chr_len - skip)
            trials = trials + 1
        if p1 + cn_size > chr_len - skip:
            p2 = chr_len - skip - 1
        else:
            p2 = p1 + cn_size
        return int(p1), int(p2)
    
def binary_search(x, array, l, r):
    # find the index i where x < array[i] and x > array[i-1]
    t = l + int((r - l)/2)
    if r > l:
        if x < array[t]:
            r = t
            return binary_search(x, array, l, r)
        else:
            l = t + 1
            return binary_search(x, array, l, r)
    return r

def get_chr(chrlen_array):
    # draw a chromosome with the chance linear to the chromosome length
    acum = []
    for i in range(len(chrlen_array)):
        if i == 0:
            acum.append(chrlen_array[i])
        else:
            acum.append(acum[i-1] + chrlen_array[i])
    # randomly draw a position from the whole genome
    x = np.random.uniform(1, acum[len(chrlen_array)-1])
    return binary_search(x, acum, 0, len(chrlen_array) - 1)

# given a nucleotide, randomly draw a new nucleotide
def get_new_nuc(nuc):
    while True:
        p = abs(int(np.random.uniform(0, 4, 1) - 0.0001))
        new_nuc = nuc_array[p]
        if nuc != new_nuc:
            return new_nuc


    
# given an array of len, find which chr and pos the p on the whole genome corresponds to
def wg2chr(chrlen, p):
    num = 0
    for l in range(len(chrlen)):
        p_ = p - chrlen[l]
        if p_ < 0:
            return num, p
        p = p_
        num = num + 1

    return "NA", -1


# given reference fasta, chrlen, snp_rate, branch length, add snvs by specifying its chr, position and change
def add_SNV(chrlen, ref, snv_rate, length):
    ret_ref = [row[:] for row in ref]
    for ale in range(len(chrlen)):
        # decide how many snvs to be added according to snv_rate and length
        mean = snv_rate * length
        # make it to be a Poisson distribution
        num = np.random.poisson(mean, 1)
        total_l = 0
        for chr in range(len(chrlen[ale])):
            total_l = total_l + chrlen[ale][chr]
        # random over the accumulation of chromosomes, then figure out which chromomome it is
        snv_n = 0
        # dictionary for the ps on the whole genome to avoid duplicated draw
        snv_ps = []
        # to store all the snvs
        snvs = []
        while snv_n < num:
            # draw a snv
            p_tmp = np.random.uniform(1, total_l, 1)
            p = int(p_tmp[0])
            if p not in snv_ps:
                chr, pos = wg2chr(chrlen[ale], p)
                print(chr, pos, ret_ref[ale][chr][pos])
                if ret_ref[ale][chr][pos] != 'N':
                    # add to dictionary
                    snv_ps.append(p)
                    # randomly find a non-N nucleotide and randomly find an alternative for SNV
                    nuc = ret_ref[ale][chr][pos]
                    new_nuc = get_new_nuc(nuc)
                    snvs.append(MySNV(ale, chr, pos, nuc, new_nuc))
                    ret_ref[ale][chr][pos] = new_nuc
        return ret_ref, snvs


def add_CN(chrlen, cn_num, del_rate, min_cn_size, exp_theta, amp_p, corres, CN_LIST_ID):
    # for each branch, the copy number change happens on only one allele
    CN_array = []
    # a variable used only for fixing CN, for test
    new_chrlen = [row[:] for row in chrlen]
    new_corres = [row[:] for row in corres]
    if random == 0:
        CN_Tot = 9
    else:
        CN_Tot = int(np.random.poisson(cn_num, 1))
    for i in range(CN_Tot):
        # allele
        if random == 0:
            CN_Ale = allele_list[i] 
        else:
            CN_Ale = np.random.binomial(1, 0.5)
        # deletion versus amplification 
        if random == 0:
            CN_Del = if_del_list[i]
        else:
            CN_Del = np.random.binomial(1, del_rate)
        CN_chromosome = get_chr(new_chrlen[CN_Ale])
        #print CN_Ale, CN_chromosome
        CN_p1, CN_p2 = get_range(new_chrlen[CN_Ale][CN_chromosome], min_cn_size, exp_theta, i)
        # to protect from str to int, int to str, make it not dividable by 10
        if CN_p1 % 10 == 0:
            CN_p1 = CN_p1 + 1
        if CN_p2 % 10 == 0:
            CN_p2 = CN_p2 - 1
# think about how to get the actual coordinates given the previous ones.
        CN_amp_num = 0
        #print new_chrlen
        #print "before changing:"
        #print chrlen, new_chrlen
        if CN_Del == 0:
            # get amplification copy number
            # starting from 0
            while 1:
                CN_amp_num_ = np.random.geometric(amp_p, 1)
                CN_amp_num = int(CN_amp_num_[0])
                if CN_amp_num >= 1:
                    break;
            if random == 0:
                CN_amp_num = amp_num_list[i]
            #CN_amp_num = int(np.random.geometric(amp_p, 1) - 1)
            #print CN_amp_num, CN_p2, CN_p1
            new_chrlen[CN_Ale][CN_chromosome] = new_chrlen[CN_Ale][CN_chromosome] + CN_amp_num * (CN_p2 - CN_p1)
            #print new_chrlen
        else:
            new_chrlen[CN_Ale][CN_chromosome] = new_chrlen[CN_Ale][CN_chromosome] - (CN_p2 - CN_p1)
            #print new_chrlen
        #print "after changing:"
        #print chrlen, new_chrlen
        # corresp is the corresponding interval of ref and genome
        new_CN = CN(CN_Ale, CN_Del, CN_chromosome, CN_p1, CN_p2, CN_amp_num, new_corres)
        new_corres = get_new_corres(new_CN, new_corres)
        #print_corrs(new_corres)
        CN_array.append(new_CN)
    return CN_array, new_chrlen, new_corres

def print_corrs(corres):
    # print out the correspondence to check its correctness
    for i in range(len(corres)):
        print("allele: " + str(i))
        for j in range(len(corres[i])):
            print("chromosome: " + str(j))
            for k in range(len(corres[i][j])):
                corr = corres[i][j][k]
                ref = corr.ref
                gen = corr.gen
                print("ref: " + str(ref))
                print("gen: " + str(gen))


def print_hash(hash_):
    for key in sorted(hash_):
        print(str(key) + " " + str(hash_[key]))

# given a hash, break all overlapping keys into non-overlapping ones and preserve the value for each segment, key is in start.end
def break_overlap(hash_):
    #bps = sorted(hash_)
    bps = list(hash_.keys())
    # hash that takes in the start or end breakpoint, with the value indicating the cn and status
    ret_hash = {}
    bp_hash = {}
    for i in range(len(bps)):
        bp_s, bp_e = str(bps[i]).split(".")
        bp_s = int(bp_s)
        bp_e = int(bp_e)
        bp_hash[bp_s] = 1
        bp_hash[bp_e] = 1
    bp_array = sorted(bp_hash)
    # make a pair of neighboring bps
    pair_array = []
    for i in range(len(bp_array) - 1):
        pair_array.append(".".join([str(bp_array[i]), str(bp_array[i + 1])]))
    # for each neighboring bp (j), see which intervals (i) contain it
    for i in range(len(bps)):
        cn = hash_[bps[i]]
        i_s, i_e = bps[i].split(".")
        i_s = int(i_s)
        i_e = int(i_e)
        for j in pair_array:
            # check if j is in i, if yes, add i's cn to j
            j_s, j_e = j.split(".")
            j_s = int(j_s)
            j_e = int(j_e)
            # j must be smaller than i
            if j_s >= i_s and j_e <= i_e:
                #key = float(j)
                key = j
                if key in list(ret_hash.keys()):
                    ret_hash[key] = ret_hash[key] + cn
                else:
                    ret_hash[key] = cn 
                            
    return ret_hash
        
# given the correspondence, summarize for each segment, the number of copies (for each allele), and a summary of deletion and amplification (merge neighbors if they are of the same copy number)
def get_cn_from_corres(corres, ref_len):
    cn_detail = []
    for i in range(len(corres)):
        # which allele
        cn_detail_ = []
        for j in range(len(corres[i])):
            # which chromosome 
            hash_ = {}
            #cn_detail_ = []
            for k in range(len(corres[i][j])):
                ref_se = corres[i][j][k].ref
                ref_s = ref_se[0]
                ref_e = ref_se[1]
                key = str(ref_s) + "." + str(ref_e)
                #key = float(key)
                if key in list(hash_.keys()):
                    hash_[key]= hash_[key] + 1
                else:
                    hash_[key] = 1
            # in case some key in hash_ overlap with others, need to break them into nonoverlapping ones
            hash_ = break_overlap(hash_)
            cn_detail_.append(get_cn_detail(hash_, ref_len[j]))
            #print "At allele " + str(i) + ", chromosome " + str(j)
            #print_hash(hash_)
        cn_detail.append(cn_detail_)
    #cn_summary = get_cn_summary(cn_detail)
    cn_summary = get_cn_summary_(cn_detail)
    return cn_detail, cn_summary

# given the detailed cn (abnormal ones) on each allele and each chromosome, summarize by merging alleles, giving the correct answer that the CNV detectors are trying to find
# can deal with collapsed start and end positions
def get_cn_summary_(cn_detail):
     # exchange allele and chromosome
    tmp_range = range(len(cn_detail[0]))
    cn_detail_ = []
    for i in tmp_range:
        cn_detail__ = []
        for j in range(len(cn_detail)):
            cn_detail__.append(0) 
        cn_detail_.append(cn_detail__)

    for i in range(len(cn_detail)):
        for j in range(len(cn_detail[i])):
            cn_detail_[j][i] = cn_detail[i][j]

    ret_summary = {}
    # now starts with chromosome
    for i in range(len(cn_detail_)):
        # a hash table with only the breakpoints, forgetting other info
        bp = {}
        # get the breakpoint hash
        # allele
        for j in range(len(cn_detail_[i])): 
            # chr
            for key in cn_detail_[i][j]:
                s, e = key.split(".")
                s = int(s)
                e = int(e)
                bp[s] = 1
                bp[e] = 1
        # now segment for each neighbor
        bps = sorted(bp)
        pair_bp = {}
        for k in range(len(bps) - 1):
            pair_bp[str(bps[k]) + "." + str(bps[k + 1])] = 1

        # get the copy number for each pair, this will include those that are not CN, will filtered out later on
        ret_summary_ = {}
        pair_bps = list(pair_bp.keys())
        pair_bps.sort(key=natural_keys) 
        for pair in pair_bps:
            pair_s, pair_e = pair.split(".")
            pair_s = int(pair_s)
            pair_e = int(pair_e)
            cn = 0
            tag = 0
            # see which big ones it belong to, if it belongs to nothing, there's no CNV 
            for j in range(len(cn_detail_[i])):
                # each allele get counted once and once only; if not counted, it means that allele is normal, need to add 1 there
                for key in cn_detail_[i][j]:
                    s, e = key.split(".")
                    s = int(s)
                    e = int(e)
                    if pair_s >= s and pair_e <= e:
                        # add the cn of this interval 
                        cn = cn + cn_detail_[i][j][key]
                        tag = tag + 1
            # have to enter each allele if the allele has a cnv, otherwise the cn there is 1
            if tag == 1:
                cn = cn + 1
            elif tag == 0:
                cn = 2
            if cn != 2:
                # only record CNV
                ret_summary_[pair] = cn
        ret_summary[i] = ret_summary_
    return ret_summary
            
# given the detailed cn (abnormal ones) on each allele and each chromosome, summarize by merging alleles, giving the correct answer that the CNV detectors are trying to find
def get_cn_summary(corres):
    cn_summary = {}
    corres_ = []

    # exchange allele and chromosome
    tmp_range = range(len(corres[0]))
    for i in tmp_range:
        corres__ = []
        for j in range(len(corres)):
            corres__.append(0) 
        corres_.append(corres__)

    for i in range(len(corres)):
        for j in range(len(corres[i])):
            corres_[j][i] = corres[i][j]

    for chr in range(len(corres_)):
        union_hash = {}
        # for each chromosome
        cn_summary_ = {}
        for allele in range(len(corres_[chr])):
            # conclude all breakpoints
            tmp_hash = corres_[chr][allele]
            # union the two alleles' cnv
            for key in tmp_hash:
                s, e = key.split(".")
                # make it string with allele
                # record both allele and start and end info in key (s: 0; e: 1; in the second digit after dot)
                key_ = str(s) + "." + str(allele) + str(1) 
                union_hash[key_] = ".".join([str(tmp_hash[key]), str(allele), "s"])
                key_ = str(e) + "." + str(allele) + str(0)
                union_hash[key_] = ".".join([str(tmp_hash[key]), str(allele), "e"])

        # summarize this chromosome
        #bps = sorted(union_hash)
        bps = list(union_hash.keys())
        bps.sort(key=natural_keys)
        prev_SorE = "NA"
        prev_prev_SorE = "NA"
        prev_cn = -1
        prev_prev_cn = -1
        prev_allele = -1
        prev_prev_allele = -1
        for i in range(len(bps)):
            bp = bps[i]
            cn, allele, SorE = union_hash[bp].split(".")
            cn = int(cn)
            bp = int(float(bp))
            if SorE == "s":
                # start
                if prev_SorE == "s":
                    # start for previous, two different alleles, conclude the previous one 
                    if prev_bp != bp:
                        key_ = ".".join([str(prev_bp), str(bp)])
                        cn_ = prev_cn + 1
                        cn_summary_[key_] = cn_
                elif prev_SorE == "e":
                    # can be either from the same or a different allele
                    #if prev_allele != allele:
                        # different, nothing
                    if prev_allele == allele:
                        # same, need to see the previous previous, if it is an s from a different allele, need to take that into account
                        # must have the prev prev as the previous is e
                        if prev_prev_SorE == "s" and prev_prev_allele != allele:
                            if prev_bp != bp:
                                key_ = ".".join([str(prev_bp), str(bp)])
                                cn_ = prev_prev_cn + 1
                                cn_summary_[key_] = cn_
            else:
                # end
                if prev_SorE == "e":
                    # the same, conclude the second e
                    if prev_bp != bp:
                        key_ = ".".join([str(prev_bp), str(bp)])
                        cn_ = cn + 1
                        cn_summary_[key_] = cn_
                elif prev_SorE == "s":
                    # check the previous previous bp
                    if prev_prev_SorE == "s":
                        # s ----------------
                        #        s --- e
                        # or
                        #        s ---------
                        # s ---------- e
                        if prev_bp != bp:
                            key_ = ".".join([str(prev_bp), str(bp)])
                            cn_ = prev_prev_cn + prev_cn
                            cn_summary_[key_] = cn_
                    elif prev_prev_SorE == "e":
                        # e
                        #        s --- e
                        # or
                        # e      s --- e
                        key_ = ".".join([str(prev_bp), str(bp)])
                        cn_ = cn + 1
                        cn_summary_[key_] = cn_
                    elif prev_prev_SorE == "NA":
                        key_ = ".".join([str(prev_bp), str(bp)])
                        cn_ = cn + 1
                        cn_summary_[key_] = cn_
            prev_bp = bp
            prev_prev_allele = prev_allele
            prev_allele = allele
            prev_prev_cn = prev_cn
            prev_cn = cn
            prev_prev_SorE = prev_SorE
            prev_SorE = SorE 
        cn_summary[chr] = cn_summary_
    return cn_summary
           
# human sort to get rid of the problem of having 0 at the end (string to float will miss this info)
def natural_keys(text):
    #return [ atof(c) for c in re.split(r'[+-]?([0-9]+(?:[.][0-9]*)?|[.][0-9]+)', text) ]
    return float(text)

# given a hash table listing the copies of each segment in a chromosome and an allele, as well as the ref length on this chromosome, combine those neighboring segments with the same copy number, find the gap (deletion), return a list with the segments that have CNVs, along with the copy number
def get_cn_detail(hash_, ref_len_):
    ret_dict = {}
    prev_e = -1
    rem_s = -1
    rem_e = -1
    keys_ = list(hash_.keys())
    keys_.sort(key=natural_keys)
    #for key in sorted(hash_):
    for key in keys_:
        s, e = key.split(".")
        s = int(s)
        e = int(e)
        cn = hash_[key]
        # first check if there is a gap
        if prev_e != -1 and prev_e == s:
            # continuous, no gap
            if cn != 1:
                # this is a CNV
                # check if the same as the previous one, if yes, merge
                if prev_cn == cn:
                    rem_e = e
                elif prev_cn != 1:
                    # if previous is also CNV, conclude it
                    seg = ".".join([str(rem_s), str(rem_e)])
                    ret_dict[seg] = prev_cn
                    # start a new one
                    rem_s = s
                    rem_e = e
                else:
                    # prev_cn == 2
                    rem_s = s
                    rem_e = e
            elif prev_cn != 1:
                # this is not a CNV, but the previous is
                # conclude it
                seg = ".".join([str(rem_s), str(rem_e)])
                ret_dict[seg] = prev_cn
                rem_s = -1
                rem_e = -1
        elif prev_e == -1:
            # just started, check if s = 1:
            if s != 0:
                # a gap
                rem_s = 0
                rem_e = s
                # conclude it
                seg = ".".join([str(rem_s), str(rem_e)])
                ret_dict[seg] = 0
            if cn != 1:
                # check if this is a CNV
                rem_s = s
                rem_e = e
        else:
            # there is a gap, record it
            # see first if there's any remnant
            if rem_s != -1:
                seg = ".".join([str(rem_s), str(rem_e)])
                ret_dict[seg] = prev_cn
            seg = ".".join([str(prev_e), str(s)])
            ret_dict[seg] = 0
            if cn != 1:
                rem_s = s
                rem_e = e
        prev_s = s
        prev_e = e
        prev_cn = cn
    # the last part missing
    if prev_e != ref_len_:
        seg = ".".join([str(prev_e), str(ref_len_)])
        ret_dict[seg] = 0
        # the previous previous one hasn't been done yet
        if rem_s != -1:
            seg = ".".join([str(rem_s), str(rem_e)])
            ret_dict[seg] = prev_cn 
    elif rem_e != -1:
        seg = ".".join([str(rem_s), str(rem_e)])
        ret_dict[seg] = prev_cn
    return ret_dict 

# given two positions on the genome, and whether it is a deletion, get the corresponding positions on the reference
def get_correspond_pos(p1, p2, if_del, new_corres, amp_num):
    # this is a list of correspondence
    ret_corres = []
    cnv_sz = p2 - p1
    amp_num = amp_num + 1
    #if len(new_corres) == 1:
    #    # there is only one, meaning that there have never been any CNV happened before
    #    corres = new_corres[0]
    #    r1, r2 = corres.ref
    #    g1, g2 = corres.gen
    #    if if_del == 1:
    #        # break into two
    #        # check if all delete
    #        if g2 - cnv_sz < 0:
    #            print "Not any left after deleting " + p1 + ", " + p2
    #            sys.exit(0)
    #        # p1 not included, always [a, b), including cnv ranges
    #        ret_corres.append(corres_coord(0, p1, 0, p1))
    #        ret_corres.append(corres_coord(p2, r2, p1, g2 - cnv_sz)) 
    #    else:
    #        # break into two, in which a part of ref belongs to both segments
    #        ret_corres.append(corres_coord(0, p2, 0, p2))
    #        ret_corres.append(corres_coord(p1, r2, p1, g2))
    #else:
    i_start = 0
    gen_off = 0
    # there are some segments already there
    for i in range(len(new_corres)):
        r1, r2 = new_corres[i].ref
        g1, g2 = new_corres[i].gen
        if p1 >= g1 and p1 < g2:
            # get p1's ref pos
            p1_ref = p1 - g1 + r1
            if if_del == 1:
                gen_off = - cnv_sz
                # a deletion, get p2
                if p2 <= g2:
                    # simply within this range of correspondence, no other ranges are involved
                    # made into to segments
                    # r1 ---- |  p1_ref  | ---- r2
                    # g1 ---- | p1 -- p2 | ---- g2
                    #ret_corres.append(corres_coord(r1, p1_ref - 1, g1, p1 - 1))
                    #ret_corres.append(corres_coord(p1_ref, r2, p2, g2))
                    #if p1 == g1 and p2 == g2:
                        # just delete this correspondence, not add anything, but everything following need to be updated on gen side
                    if p1 != g1 or p2 != g2:
                        if p1 == g1:
                            # notice on the genome side, it is the coordinate after the deletion, as we are looking at the correspondence of the genome after the deletion
                            ret_corres.append(corres_coord(p1_ref, r2, g1, g1 + (r2 - p1_ref))) 
                        elif p2 == g2:
                            ret_corres.append(corres_coord(r1, r2 - cnv_sz, g1, g2 - cnv_sz)) 
                        else:
                            # the middle part taken, broken into two
                            ret_corres.append(corres_coord(r1, p1_ref, g1, p1))
                            ret_corres.append(corres_coord(p1_ref + cnv_sz, r2, p1, p1 + (r2 - p1_ref - cnv_sz)))
                    # stop and record where it is
                    i_start = i + 1
                    break
                else:
                    # now there are two correspondences involved, thus the deletion is broken into two or more (depending on how many correspondences it involves) on the reference
                    # for this correspondence, there is only one correspondence left
                    # r1 ----| p1_ref 
                    # g1 --- | p1 
                    # see anything left for the header
                    if p1 > g1:
                        ret_corres.append(corres_coord(r1, p1_ref, g1, p1))
                    p2_ref = r2
                    # get the rest of the correspondences
                    i = i + 1
                    r1, r2 = new_corres[i].ref
                    g1, g2 = new_corres[i].gen 
                    while p2 > g2:
                        i = i + 1
                        r1, r2 = new_corres[i].ref
                        g1, g2 = new_corres[i].gen 
                    # all of these are deleted until the rest of the correspondence
                    # r1 ---- (deleted) ---- p2_ref -- r2
                    # g1 ---- (deleted) ---- p2 ------ g2
                    if p2 != g2:
                        p2_ref = r1 + p2 - g1 
                        ret_corres.append(corres_coord(p2_ref, r2, p1, p1 + g2 - p2))
                    i_start = i + 1
                    break
            else:
                # an amplification
                # amp_num includes the original copy, but the shift is just what was added to the genome, so minus 1
                gen_off = cnv_sz * (amp_num - 1)
                #print "g2: " + str(g2) + ", p2: " + str(p2)
                # an amplification, get p2
                if p2 <= g2:
                    # simply within this range of correspondence, no other ranges are involved
                    # made into to segments
                    # r1 ---- |  p1_ref  | ---- r2
                    # g1 ---- | p1 -- p2 | ---- g2
                    #ret_corres.append(corres_coord(r1, p1_ref - 1, g1, p1 - 1))
                    #ret_corres.append(corres_coord(p1_ref, r2, p2, g2))
                    if p1 == g1 and p2 == g2:
                        for j in range(amp_num):
                            ret_corres.append(corres_coord(r1, r2, g1 + j * cnv_sz, g2 + j * cnv_sz))
                        # just delete this correspondence, not add anything, but everything following need to be updated on gen side
                    else:
                        if p1 == g1:
                            # notice on the genome side, it is the coordinate after the deletion, as we are looking at the correspondence of the genome after the deletion
                            for j in range(amp_num):
                                ret_corres.append(corres_coord(r1, r1 + cnv_sz, g1 + j * cnv_sz, g1 + (j + 1) * cnv_sz)) 
                            ret_corres.append(corres_coord(r1 + cnv_sz, r2, g1 + amp_num * cnv_sz, g1 + amp_num * cnv_sz + (r2 - p1_ref))) 
                        elif p2 == g2:
                            ret_corres.append(corres_coord(r1, r2 - cnv_sz, g1, g2 - cnv_sz)) 
                            for j in range(amp_num):
                                ret_corres.append(corres_coord(r2 - cnv_sz, r2, g2 - cnv_sz + j * cnv_sz, g2 - cnv_sz + (j + 1) * cnv_sz))
                        else:
                            # the middle part taken, broken into two
                            ret_corres.append(corres_coord(r1, p1_ref, g1, p1))
                            for j in range(amp_num):
                                ret_corres.append(corres_coord(p1_ref, p1_ref + cnv_sz, p1 + j * cnv_sz, p1 + (j + 1) * cnv_sz))
                            ret_corres.append(corres_coord(p1_ref + cnv_sz, r2, p1 + amp_num * cnv_sz, p1 + amp_num * cnv_sz + (r2 - p1_ref - cnv_sz)))
                    # stop and record where it is
                    i_start = i + 1
                    break
                else:
                    # now there are two correspondences involved, thus the deletion is broken into two or more (depending on how many correspondences it involves) on the reference
                    # for this correspondence, there is only one correspondence left
                    # r1 ----| p1_ref 
                    # g1 --- | p1 
                    # see anything left for the header
                    current_i = i
                    if p1 > g1:
                        ret_corres.append(corres_coord(r1, p1_ref, g1, p1))
                        end = p1
                    elif p1 == g1:
                        end = g1
                    # to keep this value as in the for loop, it will change
                    r2_ori = r2
                    # repeated amp_num times
                    for j in range(amp_num):
                        i = current_i
                        ret_corres.append(corres_coord(p1_ref, r2_ori, end, end + r2_ori - p1_ref))
                        end = end + r2_ori - p1_ref

                        i = i + 1
                        r1, r2 = new_corres[i].ref
                        g1, g2 = new_corres[i].gen
                        
                        while p2 > g2:
                            # deal with the last
                            ret_corres.append(corres_coord(r1, r2, end, end + r2 - r1))
                            end = end + r2 - r1
                            i = i + 1
                            r1, r2 = new_corres[i].ref
                            g1, g2 = new_corres[i].gen

                        # get the residual
                        ret_corres.append(corres_coord(r1, r1 + p2 - g1, end, end + p2 - g1))
                        end = end + p2 - g1

                    # get the residual for the rest of this segment
                    ret_corres.append(corres_coord(r1 + p2 - g1, r2, end, end + r2 - r1 - p2 + g1))
                    i_start = i + 1
                    break
        elif p1 >= g2:
            # not yet
            ret_corres.append(corres_coord(r1, r2, g1, g2))
        # for the rest, p1 < g1, has been taken care of in the while loops above and the following for loop
    # see if there are left to shift on the gen side
    if i_start != 0 and i_start != len(new_corres):
        #print "i_start" + str(i_start)
        for j in range(len(new_corres) - i_start):
            k = j + i_start
            r1, r2 = new_corres[k].ref
            g1, g2 = new_corres[k].gen
            ret_corres.append(corres_coord(r1, r2, g1 + gen_off, g2 + gen_off))

    return ret_corres
        
 
def add_whole_amp(chrlen, whole_amp_rate, whole_amp_num, corres, amp_num_geo_par):
    new_chrlen = [row[:] for row in chrlen]
    new_corres = [row[:] for row in corres]
    # like CN
    wholeamp = []
    for i in range(len(chrlen)):
        # allele
        for j in range(len(chrlen[i])):
            # chromosome
            random_ = np.random.uniform(0.0,1.0)
            if random_ < whole_amp_rate:
                # chosen to be amplified
                # get the amplification number
                # geometric distribution's mean is 1/p = 1.
                amp_num = whole_amp_num * np.random.geometric(amp_num_geo_par)
                while amp_num < 1:
                    amp_num = whole_amp_num * np.random.geometric(amp_num_geo_par)
                new_CN = CN(i, 0, j, 0, new_chrlen[i][j], amp_num, new_corres)
                new_chrlen[i][j] = new_chrlen[i][j] * (amp_num + 1)
                new_corres = get_new_corres(new_CN, new_corres) 
                wholeamp.append(new_CN)
    return wholeamp, new_chrlen, new_corres

def get_new_corres(CN, corres):
    # generate the new correspondence between reference and each allele in the genome 
    ale = CN.CN_Ale
    if_del = CN.CN_Del
    chr = CN.CN_chromosome
    p1 = CN.CN_p1
    p2 = CN.CN_p2
    amp_num = CN.CN_amp_num
    new_corres = [row[:] for row in corres]
    # where the CNA occurs
    c = corres[ale][chr]
    # find where p1 and p2 correspond to on the reference
    ret_corres = get_correspond_pos(p1, p2, if_del, c, amp_num)
    # replace the original new_corres[ale][chr]
    new_corres[ale][chr] = ret_corres
    # return this
    return new_corres
    
    

def is_in(a, mytuple):
    if float(a)>float(mytuple[0]) and float(a)<=float(mytuple[1]):
        return True
    return False

def print_chr_len(chrlen_array):
    print("chr len:")
    print(chrlen_array)
    return ""

# root_mult is the multiplier of the mean CNV on root branch than those on the leaves
# whole_amp: if there's whole chromosome amplification
# whole_amp_rate: rate of an allele on a chromosome chosen to be amplified. 
# whole_amp_num: the mean of the number of copies added
def gen_tree(n, Beta, Alpha, Delta, Output, cn_num, del_rate, min_cn_size, exp_theta, amp_p, template_ref, outfile, fa_prefix, snv_rate, root_mult, whole_amp, whole_amp_rate, whole_amp_num, amp_num_geo_par):
    #n = 4
    #Beta = 0.5
    #Alpha = 0.5
    #Delta = 0
    #Output = "test"
    #cn_num = 1
    #del_rate = 0.5
    #min_cn_size = 200000
    ## exponential distribution
    ## smaller exp_theta means larger chance to get larger CNV 
    #exp_theta = 0.000001
    ## geometric distribution
    ## like simulated annealing, lower amp_p means larger chance to get large CN amp
    #amp_p = 0.5
    ##template_ref = "ref.fasta"
    #template_ref = "/home1/03626/xfan/reference/hg19.fa"
    #outfile = "/work/03626/xfan/lonestar/std.out"
    #fa_prefix = "/work/03626/xfan/lonestar/ref"
    
    ref_array = []
    chr_name_array = []
    chr_sz = []
    #n = int(raw_input("n:"))
    #Beta = float(raw_input("beta:"))
    #Alpha = float(raw_input("alpha:"))
    #Delta = float(raw_input("delta:"))
    #Output = raw_input("output file:")
    #cn_num = int(raw_input("mean copy number:"))
    #del_rate = float(raw_input("deletion rate [0, 1]:"))
    #min_cn_size = int(raw_input("minimum copy number size, recommend > 2000000:"))
    #exp_theta = float(raw_input("parameter for copy number size:"))
    #amp_p = float(raw_input("parameter for amplification allele #:"))
    #template_ref = raw_input("template fasta file:")
    #outfile = raw_input("Output file name:")
    #fa_f_prefix = raw_input("fasta prefix:")
    
    f = open(outfile, "w")
    
    
    
    
    
    #n= int(n)
    #Alpha = float(Alpha)
    #Beta = float(Beta)
    #Delta = float(Delta)
    # add a root (node 0) to the tree
    # edge length (there are at most 2*n - 1))
    #           root
    #            | CN0
    #          node 0
    #        / CN1   \ CN2
    #    node 1    node 2
    ti = np.random.exponential(1,2*n-1)
    #print len(ti)
    Ui = np.random.uniform(0.0,1.0,n-1)
    Vi = np.random.uniform(0.0,1.0,n-1)
    Di = np.random.uniform(0.0,1.0,n-1)
    Bi = np.random.beta(float(Alpha+1),float(Beta+1),n-1)
    
    #Normalizing the branch lengths
    summation = 0
    for t in ti:
        summation += t
    
    for T in range(0,len(ti)):
        ti[T]=float(ti[T])/float(summation)
    
    #print ti
    
    
    #Contructing the phylogeny
    # by default chromosome size
    # from hg19, Navin's 2012 paper
    #chr_sz = [249250621, 243199373, 198022430, 191154276, 180915260, 171115067, 159138663, 146364022, 141213431, 135534747, 135006516, 133851895, 115169878, 107349540, 102531392, 90354753, 81195210, 78077248, 59128983, 63025520, 48129895, 51304566, 155270560, 59373566]
    
    # root is the node before node 0 in tree
    
    root=MyNode("0: [0,1]")
    root.tuple=[0,1]
    ref_array, chr_name_array, chr_sz = init_ref(template_ref)
    chr_sz1 = []
    # data structure for corresponding coordinates for calculating the actual CNV on reference
    # copy so that the two arrays of allele length are independent
    corres2 = []
    for i in chr_sz:
        chr_sz1.append(i)
        corres = corres_coord(0, i, 0, i)
        corres2.append([corres])
        # each corres contains two alleles, each alleles contains all chromosomes, each chromosome contains a list of corres_coord data struture, which has the four tuple of ref1, ref2, gen1, gen2
    root.chrlen=[chr_sz1, chr_sz1]
    root.corres = [corres2, corres2]
    #print chr_sz
    root.id = -1
    
    Tree = []
    Tree.append(MyNode("0: [0,1]"))
    Tree[0].tuple=[0,1]
    Tree[0].id = 0
    CN_LIST_ID = 0
    # whole chromosome amplification
    if whole_amp == 1:
        Tree[0].cn, Tree[0].chrlen, Tree[0].corres = add_whole_amp(root.chrlen, whole_amp_rate, whole_amp_num, root.corres, amp_num_geo_par)
    # assume most of the CN happens on the root branch
        cn_array2, Tree[0].chrlen, Tree[0].corres = add_CN(Tree[0].chrlen, (cn_num * root_mult), del_rate, min_cn_size, exp_theta, amp_p, Tree[0].corres, CN_LIST_ID)
        for x in cn_array2:
            Tree[0].cn.append(x)
    else:
        Tree[0].cn, Tree[0].chrlen, Tree[0].corres = add_CN(root.chrlen, (cn_num * root_mult), del_rate, min_cn_size, exp_theta, amp_p, root.corres, CN_LIST_ID)
    Tree[0].cn_detail, Tree[0].cn_summary = get_cn_from_corres(Tree[0].corres, chr_sz)
    #print "Node 0:"
    #print Tree[0].chrlen
    Tree[0].parent=root
    Tree[0].edge_length = np.random.exponential(1,1)
    
    # update the reference on the node
    #Tree[0].ref = gen_ref(ref_array, Tree[0].cn)
    #tmp_ref = gen_ref(ref_array, Tree[0].cn)
    # memory issue, write it to a file
    #fa_f_prefix = fa_prefix + str(0) + "_"
    #write_ref(tmp_ref, chr_name_array, fa_f_prefix)
    #Tree[0].ref, Tree[0].snvs = add_SNV(Tree[0].chrlen, Tree[0].ref, snv_rate, Tree[0].edge_length)
    
    Tree.append(MyNode(str(1)+":[0,"+"{0:.2f}".format(Bi[0])+"]"+","+"{0:.4f}".format(ti[0])))
    Tree.append(MyNode(str(2)+":["+"{0:.2f}".format(Bi[0])+",1]"+","+"{0:.4f}".format(ti[1])))
    # add copy number
    Tree[1].cn, Tree[1].chrlen, Tree[1].corres = add_CN(Tree[0].chrlen, cn_num, del_rate, min_cn_size, exp_theta, amp_p, Tree[0].corres, CN_LIST_ID)
    Tree[1].cn_detail, Tree[1].cn_summary = get_cn_from_corres(Tree[1].corres, chr_sz)
    #print "Node 1:"
    #print Tree[1].chrlen
    Tree[2].cn, Tree[2].chrlen, Tree[2].corres = add_CN(Tree[0].chrlen, cn_num, del_rate, min_cn_size, exp_theta, amp_p, Tree[0].corres, CN_LIST_ID)
    Tree[2].cn_detail, Tree[2].cn_summary = get_cn_from_corres(Tree[2].corres, chr_sz)
    #print "Node 2:"
    #print Tree[2].chrlen
    
    # update the reference
    #Tree[1].ref = gen_ref(Tree[0].ref, Tree[1].cn)
    #Tree[2].ref = gen_ref(Tree[0].ref, Tree[2].cn)
    # memory issue. at one time at most 2.5 references, each is 6gb (2 alleles). 
    #parent_ref = read_ref(fa_prefix + str(0) + "_")
    #tmp_ref = gen_ref(parent_ref, Tree[1].cn)
    #fa_f_prefix = fa_prefix + str(1) + "_"
    #write_ref(tmp_ref, chr_name_array, fa_f_prefix)
    #tmp_ref = gen_ref(parent_ref, Tree[2].cn)
    #fa_f_prefix = fa_prefix + str(2) + "_"
    #write_ref(tmp_ref, chr_name_array, fa_f_prefix)
    
    Tree[1].parent=Tree[0]
    Tree[2].parent=Tree[0]
    # set parent ID
    Tree[1].parentID = 0
    Tree[2].parentID = 0
    Tree[1].id = 1
    Tree[2].id = 2
    Tree[1].tuple=[0,Bi[0]]
    Tree[2].tuple=[Bi[0],1]
    Tree[1].edge_length = ti[0]
    Tree[2].edge_length = ti[1]
    #Tree[1].ref, Tree[1].snvs = add_SNV(Tree[1].chrlen, Tree[1].ref, snv_rate, Tree[1].edge_length)
    #Tree[2].ref, Tree[2].snvs = add_SNV(Tree[2].chrlen, Tree[2].ref, snv_rate, Tree[2].edge_length)
    
    node_number=2
    j=1
    
    while j<n-1:
        if Vi[j] < Delta :
            for tr in Tree:
                if tr.is_leaf and is_in(Di[j], tr.getTuple()):
                    if (not tr.is_dead):
                        tr.name = tr.name+"*"
                    tr.setDead()
                    break
        else:
            for tree in Tree:
                if tree.is_leaf and is_in(Ui[j], tree.getTuple()) and (not tree.is_dead) :
                    # get the reference of this node, as it is a parent of the following two
                    this_id = tree.getID()
                    #parent_ref = read_ref(fa_prefix + str(this_id) + "_")

                    #print "the node from " + str(node_number + 1) + " to " + str(node_number+2) + "s' parent id: " + str(tree.getID())
                    a,b = tree.getTuple()
                    node_number+=2
                    #Two new children are born here
                    middle = float(Bi[j])*float((float(b)-float(a)))+float(a)
                    Tree.append(MyNode(str(node_number-1)+":["+"{0:.4f}".format(a)+","+"{0:.4f}".format(middle)+"]"+","+"{0:.4f}".format(ti[node_number-1]), parent=tree))
                    Tree.append(MyNode(str(node_number)+":["+"{0:.4f}".format(middle)+","+"{0:.4f}".format(b)+"]"+","+"{0:.4f}".format(ti[node_number]), parent=tree))
    
                    #The new intervals are assigned here
                    Tree[node_number-1].tuple=[a,middle]
                    Tree[node_number].tuple=[middle,b]
                    Tree[node_number-1].edge_length = ti[node_number-1]
                    Tree[node_number].edge_length = ti[node_number]
                    #add copy number
                    this_chrlen = tree.chrlen[:]
                    #print this_chrlen
                    #print node_number, tree.getID()
                    #print "node " + str(node_number - 1)
                    Tree[node_number-1].cn, Tree[node_number-1].chrlen, Tree[node_number-1].corres = add_CN(this_chrlen, cn_num, del_rate, min_cn_size, exp_theta, amp_p, tree.corres, CN_LIST_ID)
                    Tree[node_number-1].cn_detail, Tree[node_number-1].cn_summary = get_cn_from_corres(Tree[node_number-1].corres, chr_sz)
                    this_chrlen = tree.chrlen[:]
                    #print this_chrlen
                    #print node_number, tree.getID()
                    #print "node " + str(node_number)
                    Tree[node_number].cn, Tree[node_number].chrlen, Tree[node_number].corres = add_CN(this_chrlen, cn_num, del_rate, min_cn_size, exp_theta, amp_p, tree.corres, CN_LIST_ID)
                    Tree[node_number].cn_detail, Tree[node_number].cn_summary = get_cn_from_corres(Tree[node_number].corres, chr_sz)
                    this_chrlen = tree.chrlen[:]
                    #print this_chrlen
                    #print node_number, tree.getID()
    
                    # add reference
                    # memory issue
                    #Tree[node_number-1].ref = gen_ref(tree.ref, Tree[node_number-1].cn) 
                    #Tree[node_number].ref = gen_ref(tree.ref, Tree[node_number].cn) 
                    # now do not calculate the ref anyway, as it takes lots of hard disk space. Just get the tree with cn, then at the leaf, trace back all the cns up to root, and apply it to each leaf. This will solve both the memory and hard disk issue. 
                    #tmp_ref = gen_ref(parent_ref, Tree[node_number-1].cn)
                    #fa_f_prefix = fa_prefix + str(node_number-1) + "_"
                    #write_ref(tmp_ref, chr_name_array, fa_f_prefix)
                    #tmp_ref = gen_ref(parent_ref, Tree[node_number].cn)
                    #fa_f_prefix = fa_prefix + str(node_number) + "_"
                    #write_ref(tmp_ref, chr_name_array, fa_f_prefix)

                    # set parent id
                    Tree[node_number].parentID = this_id
                    Tree[node_number-1].parentID = this_id
                    # add snvs
                    #Tree[node_number-1].ref, Tree[node_number-1].snvs = add_SNV(Tree[node_number-1].chrlen, Tree[node_number-1].ref, snv_rate, Tree[node_number-1].edge_length)
                    #Tree[node_number].ref, Tree[node_number].snvs = add_SNV(Tree[node_number].chrlen, Tree[node_number].ref, snv_rate, Tree[node_number].edge_length)
    
                    # set id
                    Tree[node_number-1].id = node_number - 1
                    Tree[node_number].id = node_number
    
                    break
    
        j+=1
    
    #Changing names of the leaves
    #leaf_name=0
    #for nd in Tree:
    #    if nd.is_leaf:
    #        nd.name = leaf_name
    #        leaf_name+=1
    
    #for pre, fill, node in RenderTree(Tree[0]):
    #    print("%s%s" % (pre, node.name))
    
    # record the chromosome length for each leaf on the tree
    leaf_chrlen = []
    # record which are leaves
    leaf_index = []
    f.write("Before the tree, chromosomomal length is " + str(root.chrlen) + "\n")
    for i in range(len(Tree)):
        f.write("node %d: \n" % i)
        f.write("    parent = %d\n" % Tree[i].parent.getID())
        f.write("    name = " + str(Tree[i].name) + "\n")
    for i in range(len(Tree)):
        if Tree[i].is_leaf:
            leaf_index.append(i)
            leaf_chrlen.append(Tree[i].chrlen)
        cn = Tree[i].cn
        f.write("node %d from %d: total CN # = %d\n" % (i, Tree[i].parent.getID(), len(cn)))
        for j in range(len(cn)):
            f.write("    copy number %d: allele: %d, is del: %d, chromosome: %d, position: [%d, %d], amplification #: %d\n" % (j, cn[j].CN_Ale, cn[j].CN_Del, cn[j].CN_chromosome, cn[j].CN_p1, cn[j].CN_p2, cn[j].CN_amp_num))
        # write the copy number summary (on the reference coordinate
        cn_summary = Tree[i].cn_summary
        for chr in sorted(cn_summary):
            f.write("At chromosome %s\n" % (chr))
            cn_summary_ = cn_summary[chr]
            for each_summary in sorted(cn_summary_):
                f.write("   %s, %d\n" % (each_summary, cn_summary_[each_summary]))
        #snvs = Tree[i].snvs
        #for j in range(len(snvs)):
            #f.write("    snv %d: chr: %d, pos: %d, ref_nuc: %s, new_nuc: %s", snvs[j].chr, snvs[j].pos, snvs[j].nuc, snvs[j].new_nuc)
            
        f.write("    " + str(Tree[i].chrlen) + "\n")
            #print_chr_len(Tree[i].chrlen)
    #RenderTreeGraph(Tree[0]).to_picture(str(Output))
    
    
    # generate reference for each leaf
    # memory issue, already written.
    #for i in range(len(Tree)):
    #    fa_f_prefix = fa_prefix + str(i) + "_"
    #    write_ref(Tree[i].ref, chr_name_array, fa_f_prefix)
    
    f.close()
    return leaf_chrlen, leaf_index, chr_name_array, Tree



# rewrite it and wrap it in gen_tree.py
#!/usr/bin/env python2
# -*- coding: utf-8 -*-
# add copy number, each branch has a copy number variation
# assume the genome is diploid, and not go aneuploid
"""
Created on Sat Mar 24 13:59:59 2018

@authors: Xian Fan and Mohammad Edrisi
Contacting email: xf2@rice.edu
"""

# -*- coding: utf-8 -*- 
from anytree import Node, RenderTree
import numpy as np
import graphviz as gv
from anytree.dotexport import RenderTreeGraph
from CN import CN
from Gen_Ref_Fa import gen_ref, init_ref, write_ref

n = 4
Beta = 0.5
Alpha = 0.5
Delta = 0
Output = "test"
cn_num = 1
del_rate = 0.5
#del_rate = 0
min_cn_size = 200000
#min_cn_size = 2000000
# exponential distribution
# smaller exp_theta means larger chance to get larger CNV 
exp_theta = 0.000001
# geometric distribution
# like simulated annealing, lower amp_p means larger chance to get large CN amp
amp_p = 0.5
#template_ref = "ref.fasta"
template_ref = "/home1/03626/xfan/reference/hg19.fa"
outfile = "/work/03626/xfan/lonestar/std.out"
fa_prefix = "/work/03626/xfan/lonestar/ref"

ref_array = []
chr_name_array = []
chr_sz = []
#n = int(raw_input("n:"))
#Beta = float(raw_input("beta:"))
#Alpha = float(raw_input("alpha:"))
#Delta = float(raw_input("delta:"))
#Output = raw_input("output file:")
#cn_num = int(raw_input("mean copy number:"))
#del_rate = float(raw_input("deletion rate [0, 1]:"))
#min_cn_size = int(raw_input("minimum copy number size, recommend > 2000000:"))
#exp_theta = float(raw_input("parameter for copy number size:"))
#amp_p = float(raw_input("parameter for amplification allele #:"))
#template_ref = raw_input("template fasta file:")
#outfile = raw_input("Output file name:")
#fa_f_prefix = raw_input("fasta prefix:")

f = open(outfile, "w")




class MyNode(Node):
    def __init__(self, name, parent=None):
        Node.__init__(self, name, parent)
        self.id = 0
        self.name = name
        self.parent=parent
        self.tuple=[]
        self.is_dead=False
        self.edge_length = 0
        # alelle length for each chromosome, root has the same as reference
        self.cn=[]
        self.chrlen=[]
        self.ref=[]
    def getTuple(self):
        return self.tuple
    def setDead(self):
        self.is_dead=True
    def getID(self):
        return self.id


def get_range(chr_len, min_cn_size, exp_theta):
    # get pos1 and pos2 for the copy number. 
    # skip the first and last "skip" nucleotides (CN does not happen in that range) 
# TODO
    skip = 0
    #skip = 5000000 
    # not try any more if cannot find a good one, just use the end
    trials = 0
    max_n = 10
    cn_size = np.random.exponential(exp_theta) + min_cn_size
    p1 = np.random.uniform(skip, chr_len - skip)
    # make sure p2 not out of boundary
    while chr_len - p1 - skip - cn_size < 0 and trials < max_n:
        p1 = np.random.uniform(skip, chr_len - skip)
        trials = trials + 1
    p2 = p1 + cn_size
    return int(p1), int(p2)

def binary_search(x, array, l, r):
    # find the index i where x < array[i] and x > array[i-1]
    t = l + int((r - l)/2)
    if r > l:
        if x < array[t]:
            r = t
            return binary_search(x, array, l, r)
        else:
            l = t + 1
            return binary_search(x, array, l, r)
    return r

def get_chr(chrlen_array):
    # draw a chromosome with the chance linear to the chromosome length
    acum = []
    for i in range(len(chrlen_array)):
        if i == 0:
            acum.append(chrlen_array[i])
        else:
            acum.append(acum[i-1] + chrlen_array[i])
    # randomly draw a position from the whole genome
    x = np.random.uniform(1, acum[len(chrlen_array)-1])
    return binary_search(x, acum, 0, len(chrlen_array) - 1)

def add_CN(chrlen, cn_num, del_rate, min_cn_size, exp_theta, amp_p):
    # for each branch, the copy number change happens on only one allele
    CN_array = []
    new_chrlen = [row[:] for row in chrlen]
    CN_Tot = np.random.poisson(cn_num, 1)
    for i in range(CN_Tot):
        # allele
        CN_Ale = np.random.binomial(1, 0.5)
        # deletion versus amplification 
        CN_Del = np.random.binomial(1, del_rate)
        CN_chromosome = get_chr(new_chrlen[CN_Ale])
        #print CN_Ale, CN_chromosome
        CN_p1, CN_p2 = get_range(new_chrlen[CN_Ale][CN_chromosome], min_cn_size, exp_theta)
        CN_amp_num = 0
        #print new_chrlen
        #print "before changing:"
        #print chrlen, new_chrlen
        if CN_Del == 0:
            # get amplification copy number
            # starting from 0
            CN_amp_num = int(np.random.geometric(amp_p, 1) - 1)
            #print CN_amp_num, CN_p2, CN_p1
            new_chrlen[CN_Ale][CN_chromosome] = new_chrlen[CN_Ale][CN_chromosome] + CN_amp_num * (CN_p2 - CN_p1)
            #print new_chrlen
        else:
            new_chrlen[CN_Ale][CN_chromosome] = new_chrlen[CN_Ale][CN_chromosome] - (CN_p2 - CN_p1)
            #print new_chrlen
        #print "after changing:"
        #print chrlen, new_chrlen
        CN_array.append(CN(CN_Ale, CN_Del, CN_chromosome, CN_p1, CN_p2, CN_amp_num))
    return CN_array, new_chrlen

def is_in(a, mytuple):
    if float(a)>float(mytuple[0]) and float(a)<=float(mytuple[1]):
        return True
    return False

def print_chr_len(chrlen_array):
    print "chr len:"
    print chrlen_array
    return ""




#n= int(n)
#Alpha = float(Alpha)
#Beta = float(Beta)
#Delta = float(Delta)
# add a root (node 0) to the tree
# edge length (there are at most 2*n - 1))
#           root
#            | CN0
#          node 0
#        / CN1   \ CN2
#    node 1    node 2
ti = np.random.exponential(1,2*n-1)
#print len(ti)
Ui = np.random.uniform(0.0,1.0,n-1)
Vi = np.random.uniform(0.0,1.0,n-1)
Di = np.random.uniform(0.0,1.0,n-1)
Bi = np.random.beta(float(Alpha+1),float(Beta+1),n-1)

#Normalizing the branch lengths
summation = 0
for t in ti:
    summation += t

for T in range(0,len(ti)):
    ti[T]=float(ti[T])/float(summation)

#print ti


#Contructing the phylogeny
# by default chromosome size
# from hg19, Navin's 2012 paper
#chr_sz = [249250621, 243199373, 198022430, 191154276, 180915260, 171115067, 159138663, 146364022, 141213431, 135534747, 135006516, 133851895, 115169878, 107349540, 102531392, 90354753, 81195210, 78077248, 59128983, 63025520, 48129895, 51304566, 155270560, 59373566]

# root is the node before node 0 in tree

root=MyNode("0: [0,1]")
root.tuple=[0,1]
ref_array, chr_name_array, chr_sz = init_ref(template_ref)
chr_sz1 = []
# copy so that the two arrays of allele length are independent
for i in chr_sz:
    chr_sz1.append(i)
root.chrlen=[chr_sz, chr_sz1]
#print chr_sz
root.id = -1

Tree = []
Tree.append(MyNode("0: [0,1]"))
Tree[0].tuple=[0,1]
Tree[0].id = 0
# assume most of the CN happens on the root branch
Tree[0].cn, Tree[0].chrlen = add_CN(root.chrlen, (cn_num * 4), del_rate, min_cn_size, exp_theta, amp_p)
#print "Node 0:"
#print Tree[0].chrlen
Tree[0].parent=root
Tree[0].edge_length = np.random.exponential(1,1)

# update the reference on the node
Tree[0].ref = gen_ref(ref_array, Tree[0].cn)

Tree.append(MyNode(str(1)+":[0,"+"{0:.2f}".format(Bi[0])+"]"+","+"{0:.4f}".format(ti[0])))
Tree.append(MyNode(str(2)+":["+"{0:.2f}".format(Bi[0])+",1]"+","+"{0:.4f}".format(ti[1])))
# add copy number
Tree[1].cn, Tree[1].chrlen = add_CN(Tree[0].chrlen, cn_num, del_rate, min_cn_size, exp_theta, amp_p)
#print "Node 1:"
#print Tree[1].chrlen
Tree[2].cn, Tree[2].chrlen = add_CN(Tree[0].chrlen, cn_num, del_rate, min_cn_size, exp_theta, amp_p)
#print "Node 2:"
#print Tree[2].chrlen

# update the reference
Tree[1].ref = gen_ref(Tree[0].ref, Tree[1].cn)
Tree[2].ref = gen_ref(Tree[0].ref, Tree[2].cn)

Tree[1].parent=Tree[0]
Tree[2].parent=Tree[0]
Tree[1].id = 1
Tree[2].id = 2
Tree[1].tuple=[0,Bi[0]]
Tree[2].tuple=[Bi[0],1]
Tree[1].edge_length = ti[0]
Tree[2].edge_length = ti[1]

node_number=2
j=1

while j<n-1:
    if Vi[j] < Delta :
        for tr in Tree:
            if tr.is_leaf and is_in(Di[j], tr.getTuple()):
                if (not tr.is_dead):
                    tr.name = tr.name+"*"
                tr.setDead()
                break
    else:
        for tree in Tree:
            if tree.is_leaf and is_in(Ui[j], tree.getTuple()) and (not tree.is_dead) :
                #print "the node from " + str(node_number + 1) + " to " + str(node_number+2) + "s' parent id: " + str(tree.getID())
                a,b = tree.getTuple()
                node_number+=2
                #Two new children are born here
                middle = float(Bi[j])*float((float(b)-float(a)))+float(a)
                Tree.append(MyNode(str(node_number-1)+":["+"{0:.4f}".format(a)+","+"{0:.4f}".format(middle)+"]"+","+"{0:.4f}".format(ti[node_number-1]), parent=tree))
                Tree.append(MyNode(str(node_number)+":["+"{0:.4f}".format(middle)+","+"{0:.4f}".format(b)+"]"+","+"{0:.4f}".format(ti[node_number]), parent=tree))

                #The new intervals are assigned here
                Tree[node_number-1].tuple=[a,middle]
                Tree[node_number].tuple=[middle,b]
                Tree[node_number-1].edge_length = ti[node_number-1]
                Tree[node_number].edge_length = ti[node_number]
                #add copy number
                this_chrlen = tree.chrlen[:]
                #print this_chrlen
                #print node_number, tree.getID()
                #print "node " + str(node_number - 1)
                Tree[node_number-1].cn, Tree[node_number-1].chrlen = add_CN(this_chrlen, cn_num, del_rate, min_cn_size, exp_theta, amp_p)
                this_chrlen = tree.chrlen[:]
                #print this_chrlen
                #print node_number, tree.getID()
                #print "node " + str(node_number)
                Tree[node_number].cn, Tree[node_number].chrlen = add_CN(this_chrlen, cn_num, del_rate, min_cn_size, exp_theta, amp_p)
                this_chrlen = tree.chrlen[:]
                #print this_chrlen
                #print node_number, tree.getID()

                # add reference
                Tree[node_number-1].ref = gen_ref(tree.ref, Tree[node_number-1].cn) 
                Tree[node_number].ref = gen_ref(tree.ref, Tree[node_number].cn) 

                # set id
                Tree[node_number-1].id = node_number - 1
                Tree[node_number].id = node_number

                break

    j+=1

#Changing names of the leaves
#leaf_name=0
#for nd in Tree:
#    if nd.is_leaf:
#        nd.name = leaf_name
#        leaf_name+=1

#for pre, fill, node in RenderTree(Tree[0]):
#    print("%s%s" % (pre, node.name))

f.write("Before the tree, chromosomomal length is " + str(root.chrlen) + "\n")
for i in range(len(Tree)):
    f.write("node %d: \n" % i)
    f.write("    parent = %d\n" % Tree[i].parent.getID())
    f.write("    name = " + str(Tree[i].name) + "\n")
for i in range(len(Tree)):
    cn = Tree[i].cn
    f.write("node %d from %d: total CN # = %d\n" % (i, Tree[i].parent.getID(), len(cn)))
    for j in range(len(cn)):
        f.write("    copy number %d: allele: %d, is del: %d, chromosome: %d, position: [%d, %d], amplification #: %d\n" % (j, cn[j].CN_Ale, cn[j].CN_Del, cn[j].CN_chromosome, cn[j].CN_p1, cn[j].CN_p2, cn[j].CN_amp_num))
    f.write("    " + str(Tree[i].chrlen) + "\n")
        #print_chr_len(Tree[i].chrlen)
#RenderTreeGraph(Tree[0]).to_picture(str(Output))


# generate reference for each leaf
for i in range(len(Tree)):
    fa_f_prefix = fa_prefix + str(i) + "_"
    write_ref(Tree[i].ref, chr_name_array, fa_f_prefix)

f.close()



# definition of copy number class: only one copy number defined. 

class CN:
    def __init__(self, CN_Ale, CN_Del, CN_chromosome, CN_p1, CN_p2, CN_amp_num, corres):
        self.CN_Ale = CN_Ale
        self.CN_Del = CN_Del
        self.CN_chromosome = CN_chromosome
        self.CN_p1 = CN_p1
        self.CN_p2 = CN_p2
        self.CN_amp_num = CN_amp_num
        self.corres = corres
    def get_CN_Ale(self):
        return self.CN_Ale
    def get_CN_Del(self):
        return self.CN_Del
    def get_CN_chromosome(self):
        return self.CN_chromosome
    def get_CN_position(self):
        #print self.CN_p1, self.CN_p2
        return self.CN_p1, self.CN_p2
    def get_CN_amp_num(self):
        return self.CN_amp_num



#!/usr/bin/python$

import argparse
import subprocess
import os
import numpy
import sys
import multiprocessing as mp
from gen_tree import gen_tree
from gen_readcount import gen_readcount
from gen_readcount import get_beta_dist
from Gen_Ref_Fa import make_fa 
from Gen_Ref_Fa import init_ref

# see how many processors I have
#NUM_OF_PROCESSES = mp.cpu_count()

def gen_reads(dir, index, leaf_index, all_chrlen, fa_prefix, Alpha, Beta, x0, y0, cov, l, window_size, u, chr_name_array):
    this_leaf_index = leaf_index[index]
    print(all_chrlen)
    print("####" + str(index))
    # each allele
    for i in range(len(all_chrlen)):
        fa_f = fa_prefix + str(this_leaf_index) + "_" + str(i + 1) + ".fa"
        # sequence the reads using wgsim
        out_fq_prefix = "leaf" + str(this_leaf_index) + "_allele" + str(i)
        out_fq1 = dir + "/" + out_fq_prefix + "_1.fq"
        out_fq2 = dir + "/" + out_fq_prefix + "_2.fq"
        args = "samtools faidx " + fa_f
        popen = subprocess.Popen(args, stdout=subprocess.PIPE, shell=True)
        popen.wait()
        #popen.terminate()
        #output = popen.stdout.read()
        #print output

        # each chromosome
        for j in range(len(all_chrlen[i])):
            print(j)
            this_chrlen = all_chrlen[i][j]
            readcounts = gen_readcount(dir, Alpha, Beta, x0, y0, cov, l, window_size, u, all_chrlen[i][j])
            # each bin
            start = 0
            end = 0
            for k in readcounts:
                this_readcount = int(k/2)
                # extract this segment of fa
                start = end + 1
                end = start + window_size - 1
                print(end)
                if end > this_chrlen:
                    end = this_chrlen
                if start > this_chrlen:
                    break
                tmp_fa_file = "_".join([fa_f, str(chr_name_array[j]), str(start), str(end)]) + ".fa" 
                # use samtools faidx to get this sequence that is to be sequenced into reads
                args = "samtools faidx " + fa_f + " " + chr_name_array[j] + ":" + str(start) + "-" + str(end) + " > " + tmp_fa_file
# TODO check how many N's 
# conclusion: Popen cannot exit. check_call can. But check_call does not work on all Ns as wgsim does not work on it. 
                print(args)
                try:
                    subprocess.check_call(args, shell=True)
                except subprocess.CalledProcessError:
                    print(e)
                    #error_out("Cannot work on '%s'" % args) 

                # check N's
                N_true = check_Ns(tmp_fa_file)

                #popen = subprocess.Popen(args, shell=False)
                # stdout=subprocess.PIPE, shell=True)
                #popen.wait()
                #popen.terminate()
                #output = popen.stdout.read()
                #print output

                # sequence the reads using wgsim
                #out_fq_prefix = "leaf" + str(this_leaf_index) + "_allele" + str(i) + "_chr" + str(j) + "_" + str(start) + "_" + str(end)
                #out_fq1 = out_fq_prefix + "_1.fq"
                #out_fq2 = out_fq_prefix + "_2.fq"
                if N_true == 1:
                    args = wgsim_dir + "wgsim -h -N " + str(this_readcount) + " -1 " + str(l) + " -2 " + str(l) + " " + tmp_fa_file + " " + out_fq1 + " " + out_fq2
                    print(args)
                    try:
                        subprocess.check_call(args, shell=True)
                    except subprocess.CalledProcessError as e:
                        print(e)
                        #error_out("Cannot work on '%s'" % args) 

                args = "rm " + tmp_fa_file 
                try:
                    subprocess.check_call(args, shell=True)
                except subprocess.CalledProcessError as e:
                    print(e)
                    #error_out("Cannot remove '%s'" % tmp_fa_file)
                #popen = subprocess.Popen(args)
                #, stdout=subprocess.PIPE, shell=True)
                #popen.wait()
                #popen.communicate()
                #output = popen.stdout.read()
                #print output
                #popen.terminate()
    print("Done with node " + str(this_leaf_index) + ", will remove it. ")
    
    ref_files = fa_prefix + str(this_leaf_index) + "_*.fa"
    args = "rm " + ref_files 
    try:
        subprocess.check_call(args, shell=True)
    except subprocess.CalledProcessError:
        print(e)
        #error_out("Cannot remove '%s'" % ref_files)


if len(sys.argv) <= 1:
    print("""
    A single cell simulator generating low coverage data. The program automatically generates a phylogenetic tree with copy number variations on the branches. On each leave of the tree, it generates the reads whose error profile, such as uneven coverage, mimics the real single cell data. 
    Usage: python main.py -t [ref.fa] -n [number_leafs] -S [wgsim_dir]
        -p (--processors)   Numbers of processors available.
        -r (--directory)    Location of simulated data. The program will remove the whole directory if it already exists. Otherwise it will create one. (default: test)
        -S (--wgsim-dir)    The directory of the binary of wgsim. It is in the same folder of this main.py. (need to specify) 
        -n (--leaf-num)     Number of the leafs. 
        -B (--Beta)         The program uses the Beta-splitting model to generate the phylogenetic tree. Specify a value between [0, 1]. (default: 0.5)
        -A (--Alpha)        The Alpha in Beta-splitting model. Specify a value between [0, 1]. The closer Alpha and Beta, the more balanced the tree. (default: 0.5).
        -D (--Delta)        The rate of a node to disappear. Specify a value between [0, 1]. If all nodes have daughter nodes, take 0. (default: 0)
        -c (--cn-num)       The average number of copy number variations to be added on a branch. (default: 1)
        -d (--del-rate)     The rate of deletion as compared to amplification. (default: 0.5)
        -m (--min-cn-size)  Minimum copy number size. (default: 200,000bp)
        -e (--exp-theta)    The parameter for the Exponential distribution for copy number size, beyond the minimum one. (default: 0.000001)
        -a (--amp-p)        The parameter for the Genometric distribution for the number of copies amplified. (default: 0.5)
        -t (--template-ref) The reference file to sequence the reads. 
        -o (--outfile)      The standard output file, will be saved in output folder, just give the file name. (default: std.out)
        -f (--fa-prefix)    The prefix of the alleles and read names. (default: ref)
        -x (--Lorenz-x)     The value on the x-axis of the point furthest from the diagonal on the Lorenz curve imitating the real coverage uneveness. (default: 0.5) 
        -y (--Lorenz-y)     The value on the y-axis of the Lorenz curve imitating the real coverage unevenness. x > y. The closer (x, y) to the diagonal, the better the coverage evenness. (default: 0.4) 
        -v (--coverage)     The average coverage of the sequence. (default: 0.02)
        -l (--readlen)      Read length for each read sequenced. (default: 35bp)
        -w (--window-size)  Within a window, the coverage is according to a Gaussian distribution. Neighboring windows' read coverage is according to a Metropolis Hasting process. (default: 200000bp)
        -u (--acceptance-rate)  The probability to accept a proposal in Metropolis Hasting. (default: 0.5)
        -k (--skip-first-step)  If the alleles for all nodes have been made, the step can be skipped. Make it 1 then. (default: 0)
        -R (--snv-rate)     The rate of the snv. snv-rate * branch-length = # snvs. (default: 1)
        -X (--multi-root)   The multiplier of the mean CNV on root. (default: 4)
        -W (--whole-amp)    If there is whole chromosome amplification, 1 as yes. (default: 1) 
        -C (--whole-amp-rate)   Whole amplification rate: rate of an allele chosen to be amplified (default: 0.2)
        -E (--whole-amp-num)    Whole amplification copy number addition, which occurs to one allele at a time. (default: 1)
        -J (--amp-num-geo-par)  Whole amplification copy number distribution (geometric distribution parameter: the smaller, the more evenly distributed). (default: 1)
        -Y (--leaf-index-range) For parallele job submission. >= min, < max leaf index will be processed. min.max (default: -1)
        -I (--leaf-ID-range) For parallele job submission. >= min, < max leaf ID will be processed. min.max (default: -1). When both -Y and -I are -1, all leaves will be processed.
        """)
    sys.exit(0)

parser = argparse.ArgumentParser(description='A copy number variation simulator for single cells.')
parser.add_argument('-p', '--processors', default=8)
parser.add_argument('-r', '--directory', default="test")
parser.add_argument('-S', '--wgsim-dir', default="")
parser.add_argument('-n', '--leaf-num', default=4)
parser.add_argument('-B', '--Beta', default=0.5)
parser.add_argument('-A', '--Alpha', default=0.5)
parser.add_argument('-D', '--Delta', default=0)
#parser.add_argument('-O', '--Output', default="test")
parser.add_argument('-c', '--cn-num', default=1)
parser.add_argument('-d', '--del-rate', default=0.5)
parser.add_argument('-m', '--min-cn-size', default=200000)
#parser.add_argument('-m', '--min-cn-size', default=10)
parser.add_argument('-e', '--exp-theta', default=0.000001)
parser.add_argument('-a', '--amp-p', default=0.5)
parser.add_argument('-t', '--template-ref', default="hg19.fa")
#parser.add_argument('-t', '--template-ref', default="chr1.fa")
parser.add_argument('-o', '--outfile', default="std.out")
parser.add_argument('-f', '--fa-prefix', default="ref")
parser.add_argument('-x', '--Lorenz-x', default=0.5)
parser.add_argument('-y', '--Lorenz-y', default=0.4)
parser.add_argument('-v', '--coverage', default=0.02)
#parser.add_argument('-v', '--coverage', default=10)
parser.add_argument('-l', '--readlen', default=35)
parser.add_argument('-w', '--window-size', default=200000)
#parser.add_argument('-w', '--window-size', default=20)
parser.add_argument('-u', '--acceptance-rate', default=0.5)
parser.add_argument('-k', '--skip-first-step', default=0)
parser.add_argument('-R', '--snv_rate', default=0)
parser.add_argument('-X', '--multi-root', default=4)
parser.add_argument('-W', '--whole-amp', default=0)
parser.add_argument('-C', '--whole-amp-rate', default=0.2)
parser.add_argument('-E', '--whole-amp-num', default=1)
parser.add_argument('-J', '--amp-num-geo-par', default=1)
parser.add_argument('-Y', '--leaf-index-range', default="-1")
parser.add_argument('-I', '--leaf-ID-range', default="-1")


args = parser.parse_args()
NUM_OF_PROCESSES = int(args.processors)
skip = int(args.skip_first_step)
dir = args.directory
save_prefix = dir + "/" + "from_first_step" 
wgsim_dir = args.wgsim_dir
n = int(args.leaf_num)
Beta = float(args.Beta)
Alpha = float(args.Alpha)
Delta = float(args.Delta)
#Output = args.Output
cn_num = int(args.cn_num)
del_rate = float(args.del_rate)
min_cn_size = int(args.min_cn_size)
exp_theta = float(args.exp_theta)
amp_p = float(args.amp_p)
template_ref = args.template_ref
outfile = dir + "/" + args.outfile
fa_prefix = dir + "/" + args.fa_prefix
x0 = float(args.Lorenz_x)
y0 = float(args.Lorenz_y)
cov = float(args.coverage)
l = int(args.readlen)
window_size = int(args.window_size)
u = float(args.acceptance_rate)
#sigma = args.sigma_for_Gaussian
snv_rate = float(args.snv_rate)
root_mult = int(args.multi_root)
whole_amp = int(args.whole_amp)
whole_amp_rate = float(args.whole_amp_rate)
whole_amp_num = int(args.whole_amp_num)
amp_num_geo_par = float(args.amp_num_geo_par)
leaf_index_range = args.leaf_index_range
index_min = 0
index_max = 10000000
if leaf_index_range != "-1":
    index_min, index_max = leaf_index_range.split(".")
    index_min = int(index_min)
    index_max = int(index_max)
leaf_ID_range = args.leaf_ID_range
leaf_index_min = 0
leaf_index_max = 10000000
if leaf_ID_range != "-1":
    leaf_index_min, leaf_index_max = leaf_ID_range.split(".")
    leaf_index_min = int(leaf_index_min)
    leaf_index_max = int(leaf_index_max)

if skip == 0: 
    if not os.path.exists(dir):
        subprocess.check_call("mkdir " + dir, shell=True)
    else:
        print(dir + " exists. Exit. ")
        sys.exit(0)

leaf_chrlen = []
leaf_index = []
chr_name_array = []

def check_Ns(file):
    N_line = 0
    total_line = 0
    with open(file, "r") as f:
        for line in f:
            total_line = total_line + 1
            if line.find('N') != -1:
                N_line = N_line + 1
    f.close()
    if N_line/float(total_line) > 0.2:
        return 0
    return 1



# Step 1. generate a phylogentic tree that has copy number alterations on branches.
# Now all CNs are in tree, not generating fa or remember it in the tree nodes.
if skip == 0:
    [leaf_chrlen, leaf_index, chr_name_array, tree] = gen_tree(n, Beta, Alpha, Delta, dir, cn_num, del_rate, min_cn_size, exp_theta, amp_p, template_ref, outfile, fa_prefix, snv_rate, root_mult, whole_amp, whole_amp_rate, whole_amp_num, amp_num_geo_par)
    numpy.save(save_prefix + ".leaf_chrlen.npy", leaf_chrlen)
    numpy.save(save_prefix + ".leaf_index.npy", leaf_index)
    numpy.save(save_prefix + ".chr_name_array.npy", chr_name_array)
    # save the tree for parallele job submission afterwards
    numpy.save(save_prefix + ".tree.npy", tree)
    #numpy.save(save_prefix + ".ref.npy", ref)
    print("Done with generating the tree. Save to npy. ")
#print leaf_index

# Step 2. Consider even coverage, use metropolis hasting to sample read count in each bin based on a given point on Lorenz curve. 

# if first step was skipped, read the previous stored file
if skip == 1:
    print("Skip the first step. Reading ")
    #abs_path = os.getcwd()
    leaf_chrlen_f = save_prefix + ".leaf_chrlen.npy"
    leaf_index_f = save_prefix + ".leaf_index.npy"
    chr_name_array_f = save_prefix + ".chr_name_array.npy"
    tree_f = save_prefix + ".tree.npy"
    #ref_f = save_prefix + ".ref.npy"
    leaf_chrlen = numpy.load(leaf_chrlen_f)
    leaf_index = numpy.load(leaf_index_f)
    chr_name_array = numpy.load(chr_name_array_f)
    tree = numpy.load(tree_f, allow_pickle=True)
    [ref, tmp_chr_name, tmp_len_chr] = init_ref(template_ref)
    #ref = numpy.load(ref_f)
    #print(leaf_chrlen_f)
    #print(leaf_index_f)
    #print(chr_name_array_f)

    # make it either making a tree, or generating the leaves
    [Alpha, Beta] = get_beta_dist(x0, y0)
    index = 0
    leaf_index_ = 0
    print("Alpha = %.2f, Beta = %.2f", Alpha, Beta)
    print("Number of processes: " + str(NUM_OF_PROCESSES))
    print("Number of leaves: " + str(n))
    # each leaf
    processes = []
    for all_chrlen in leaf_chrlen:
    #TODO add a function here to retrieve the CN and apply to reference, and write to a fa file for the two alleles.
        leaf_index_ = leaf_index[index]
        if leaf_index_range != "-1" and index < index_max and index >= index_min or leaf_ID_range != "-1" and leaf_index_ < leaf_index_max and leaf_index_ >= leaf_index_min or leaf_index_range == "-1" and leaf_ID_range == "-1": 
            make_fa(leaf_index[index], tree, ref, chr_name_array, fa_prefix)
            
            #if index % NUM_OF_PROCESSES == 0 and index != 0:
            
                # Run processes
                #for p in processes:
                #    p.start()
                
                # Exit the completed processes
                #for p in processes:
                #    p.join()
                
                # clean for the next batch
                #processes = []
        
                # clean the temp fa files
                #args = "rm " + fa_prefix + "*_*_*_*_*.fa"
                #popen = subprocess.Popen(args, stdout=subprocess.PIPE, shell=True)
                #popen.wait()
        
            processes.append(mp.Process(target=gen_reads, args=(dir, index, leaf_index, all_chrlen, fa_prefix, Alpha, Beta, x0, y0, cov, l, window_size, u, chr_name_array)))
    
        index = index + 1

    # clean the rest
    for p in processes:
        p.start()
    
    for p in processes:
        p.join()
    
    # clean up the temp fa files
    #args = "rm " + fa_prefix + "*_*_*_*_*.fa"
    #popen = subprocess.Popen(args, stdout=subprocess.PIPE, shell=True)
    #popen.wait()



import gen_tree
import numpy
import sys
import re

if len(sys.argv) <= 1:
    print("""
    Read the npy tree file, convert to newick file without branch length. 
    Usage: python
    """ + sys.argv[0] + """ [tree.npy]""")
    sys.exit(0)

class mynode:
    def __init__(self, ID, parent=None):
        self.ID = ID
        # ID
        self.parent = parent
        self.children = []

# step 2. convert to newick format
def convert2newick(Tree, str_, i):
    # stop
    if i >= len(Tree):
        return str_
    node = Tree[i]
    parent = node.parent
    children = node.children
    # get rid of the node itself
    c_arr = []
    for c in children:
        if c != str(i):
            c_arr.append(c)
    children = c_arr

    if len(children) >= 1:
        #print children
        # prepare for where to replace
        splitted = re.split(r'(\d+)', str_)
        i_ = -1
        for i_ in range(len(splitted)):
            if str(splitted[i_]) == str(i):
                break
        # now make a new string to replace
        new_substr = "(" + "," . join(children) + ")"
        #print str(splitted[i_]) + " is to be replaced by " + new_substr
        # concatenate
        splitted[i_] = new_substr + splitted[i_]
        str_ = "".join(splitted)
        str_ = convert2newick(Tree, str_, i + 1)
    else:
        str_ = convert2newick(Tree, str_, i + 1)
    return str_
        
        
    
tree_file = sys.argv[1]
tree = numpy.load(tree_file, allow_pickle=True)

# a new tree with a simplified structure
Tree = []

# step 1, read the npy file and put them in MyNode
for i in range(len(tree)):
    node = tree[i]
    parent = node.parentID
    ID = i
    Tree.append(mynode(ID, parent))
    # update the children of its parent
    Tree[parent].children.append(str(ID))

str_ = convert2newick(Tree, "(0)", 0)
#print str_[1:-1]
print str_






#!/usr/bin/python$
# __main__.py

import argparse
import subprocess
import os
#import numpy
import sys
#import multiprocessing as mp
from gen_tree_overlappingCNA import gen_tree
#from gen_readcount import gen_readcount
#from gen_readcount import get_beta_dist
#from Gen_Ref_Fa import make_fa, make_fa_wABs 
#from Gen_Ref_Fa import init_ref

# fixed bulk uniformity
x0_bulk = 0.5 
y0_bulk = 0.38 

# see how many processors I have
#NUM_OF_PROCESSES = mp.cpu_count()

# for both bulk and single-cell sampling purposes
def gen_reads(dir, index, leaf_index, all_chrlen, fa_prefix, Alpha, Beta, x0, y0, cov, l, window_size, u, chr_name_array, cell_i, original_level, bulk_or_sc):
    this_leaf_index = leaf_index[index]
    print(all_chrlen)
    print("####" + str(index))
    # each allele
    for i in range(len(all_chrlen)):
        fa_f = fa_prefix + str(this_leaf_index) + "_" + str(i + 1) + ".fa"
        # sequence the reads using wgsim
        #out_fq_prefix = "leaf" + str(this_leaf_index) + "_cell" + str(cell_i) + "_allele" + str(i)
        out_fq_prefix = "level" + original_level + "_node" + str(this_leaf_index) + "_cell" + str(cell_i) + "_allele" + str(i)

        if bulk_or_sc == "bulk":
            out_fq_prefix = "level" + original_level + "_node" + str(this_leaf_index) + "_allele" + str(i)

        out_fq1 = dir + "/" + out_fq_prefix + "_1.fq"
        out_fq2 = dir + "/" + out_fq_prefix + "_2.fq"
        args = "samtools faidx " + fa_f
        popen = subprocess.Popen(args, stdout=subprocess.PIPE, shell=True)
        popen.wait()
        #popen.terminate()
        #output = popen.stdout.read()
        #print output

        # each chromosome
        for j in range(len(all_chrlen[i])):
            print(j)
            this_chrlen = all_chrlen[i][j]
            readcounts = gen_readcount(dir, Alpha, Beta, x0, y0, cov, l, window_size, u, all_chrlen[i][j])
            # each bin
            start = 0
            end = 0
            for k in readcounts:
                this_readcount = int(k/2)
                # extract this segment of fa
                start = end + 1
                end = start + window_size - 1
                print(end)
                if end > this_chrlen:
                    end = this_chrlen
                if start > this_chrlen:
                    break
                tmp_fa_file = "_".join([fa_f, str(chr_name_array[j]), str(start), str(end)]) + ".fa" 
                # use samtools faidx to get this sequence that is to be sequenced into reads
                args = "samtools faidx " + fa_f + " " + chr_name_array[j] + ":" + str(start) + "-" + str(end) + " > " + tmp_fa_file
# TODO check how many N's 
# conclusion: Popen cannot exit. check_call can. But check_call does not work on all Ns as wgsim does not work on it. 
                print(args)
                try:
                    subprocess.check_call(args, shell=True)
                except subprocess.CalledProcessError:
                    error_out("Cannot work on '%s'" % args) 

                # check N's
                N_true = check_Ns(tmp_fa_file)

                if N_true == 1:
                    args = wgsim_dir + "wgsim -h -N " + str(this_readcount) + " -1 " + str(l) + " -2 " + str(l) + " " + tmp_fa_file + " " + out_fq1 + " " + out_fq2
                    print(args)
                    try:
                        subprocess.check_call(args, shell=True)
                    except subprocess.CalledProcessError:
                        error_out("Cannot work on '%s'" % args) 

                args = "rm " + tmp_fa_file 
                try:
                    subprocess.check_call(args, shell=True)
                except subprocess.CalledProcessError:
                    error_out("Cannot remove '%s'" % tmp_fa_file)
                #popen = subprocess.Popen(args)
                #, stdout=subprocess.PIPE, shell=True)
                #popen.wait()
                #popen.communicate()
                #output = popen.stdout.read()
                #print output
                #popen.terminate()
    if bulk_or_sc == "bulk":
        print("Done with node " + str(this_leaf_index) + " for bulk sequencing. ")
    else:
        print("Done with node " + str(this_leaf_index) + " cell " + str(cell_i))

    
def main():
    if len(sys.argv) <= 1:
        print("""
        A single cell simulator generating low coverage data. The program automatically generates a phylogenetic tree with copy number variations on the branches. On each leave of the tree, it generates the reads whose error profile, such as uneven coverage, mimics the real single cell data. 
        Usage: python main.py -t [ref.fa] -n [number_leafs] -S [wgsim_dir]
            -p (--processors)   Numbers of processors available.
            -r (--directory)    Location of simulated data. The program will remove the whole directory if it already exists. Otherwise it will create one. (default: test)
            -S (--wgsim-dir)    The directory of the binary of wgsim. It is in the same folder of this main.py. (need to specify) 
            -n (--cell-num)     Number of the cells on a level of interest. Always greater than -F treewidth. Treewidth controls the total number of clones whereas cell-num controls the total number of cells sequenced at a certain tree depth. (default: 8)
            -B (--Beta)         The program uses the Beta-splitting model to generate the phylogenetic tree. Specify a value between [0, 1]. (default: 0.5)
            -A (--Alpha)        The Alpha in Beta-splitting model. Specify a value between [0, 1]. The closer Alpha and Beta, the more balanced the tree. (default: 0.5).
            -D (--Delta)        The rate of a node to disappear. Specify a value between [0, 1]. If all nodes have daughter nodes, take 0. (default: 0)
            -G (--treedepth)    The mean of the tree depth distribution. The final tree depth will be sampled from a Gaussian with this mean and a fixed standard deviation. (default: 4 counting from the first cancer cell)
            -K (--treedepthsigma)	The standard deviation of the tree depth distribution. To get exactly the tree depth defined by -F, use a very small standard deviation, e.g., 0.0001. (default: 0.5)
            -F (--treewidth)    The mean of the tree width distribution. The final tree width will be sampled from a Gaussian with this mean and a fixed standard deviation. (default: 8)
            -H (--treewidthsigma)	The standard deviation of the tree width distribution. To get exactly the tree width defined by -F, use a very small standard deviation, e.g., 0.0001. (default: 0.5)
            -c (--cn-num)       The average number of copy number variations to be added on a branch. (default: 1)
            -d (--del-rate)     The rate of deletion as compared to amplification. (default: 0.5)
            -m (--min-cn-size)  Minimum copy number size. (default: 200,000bp)
            -e (--exp-theta)    The parameter for the Exponential distribution for copy number size, beyond the minimum one. (default: 0.000001)
            -a (--amp-p)        The parameter for the Genometric distribution for the number of copies amplified. (default: 0.5)
            -t (--template-ref) The reference file to sequence the reads. 
            -o (--outfile)      The standard output file, will be saved in output folder, just give the file name. (default: std.out)
            -f (--fa-prefix)    The prefix of the alleles and read names. (default: ref)
            -x (--Lorenz-x)     The value on the x-axis of the point furthest from the diagonal on the Lorenz curve imitating the real coverage uneveness. (default: 0.5) 
            -y (--Lorenz-y)     The value on the y-axis of the Lorenz curve imitating the real coverage unevenness. x > y. The closer (x, y) to the diagonal, the better the coverage evenness. (default: 0.4) 
            -v (--coverage)     The average coverage of the sequence. (default: 0.02)
            -l (--readlen)      Read length for each read sequenced. (default: 35bp)
            -w (--window-size)  Within a window, the coverage is according to a Gaussian distribution. Neighboring windows' read coverage is according to a Metropolis Hasting process. (default: 200000bp)
            -u (--acceptance-rate)  The probability to accept a proposal in Metropolis Hasting. (default: 0.5)
            -k (--skip-first-step)  If the alleles for all nodes have been made, the step can be skipped. Make it 1 then. (default: 0)
            -R (--snv-rate)     The rate of the snv. snv-rate * branch-length = # snvs. (default: 1)
            -X (--multi-root)   The multiplier of the mean CNV on root. (default: 4)
            -W (--whole-amp)    If there is whole chromosome amplification, 1 as yes. (default: 1) 
            -C (--whole-amp-rate)   Whole amplification rate: rate of an allele chosen to be amplified (default: 0.2)
            -E (--whole-amp-num)    Whole amplification copy number addition, which occurs to one allele at a time. (default: 1)
            -J (--amp-num-geo-par)  Whole amplification copy number distribution (geometric distribution parameter: the smaller, the more evenly distributed). (default: 1)
            -Y (--leaf-index-range) For parallele job submission. >= min, < max leaf index will be processed. min.max. This counts leaf nodes from 0. (default: -1)
            -L (--levels)	This is for both tree inference and sampling ancestral nodes. If the user is interested in sequencing nodes on multiple levels, user can specify the levels of interest by separating them by semicolon. The first tumor cell/clone under the trunk has level 1. If counting from the bottom (leaf) of the tree, use minus before the number. For example, -1 is the leaf level. The range of the level should be within [-depth, depth]. Users can specify desired levels according to -G to know which levels are available. If that is the case, use a very small -K to make sure the depth is not smaller than the biggest level you specify. (default: -1)
            -U (--bulk-levels)	The levels of the bulk sequencing separated by semicolon. The definition of the levels is the same as in -L. The default for this option is NA, meaning no bulk sequencing. 	
            -V (--cov-bulk)	The coverage of the bulk sequencing. The same for all levels. This parameter is needed when -U is identified. (default: 30) 	
            -M (--single-cell-per-node)	If this is on, each node represents one cell and there is no clonality in the node. In this case tree_width will be the same as n (leaf num). 1 is on. (default: 0)
            """)
        sys.exit(0)
    
    parser = argparse.ArgumentParser(description='SCSim: A simulator for simulating cancer phylogenetic tree with copy number variations and single nucleotide variations for single cells and bulk sequencing.')
    parser.add_argument('-p', '--processors', default=4)
    parser.add_argument('-r', '--directory', default="test")
    parser.add_argument('-S', '--wgsim-dir', default="")
    parser.add_argument('-n', '--leaf-num', default=8)
    parser.add_argument('-B', '--Beta', default=0.5)
    parser.add_argument('-A', '--Alpha', default=0.5)
    parser.add_argument('-D', '--Delta', default=0)
    # add the width to the tree
    parser.add_argument('-F', '--treewidth', default=8)
    # add the depth to the tree
    parser.add_argument('-G', '--treedepth', default=4)
    parser.add_argument('-H', '--treewidthsigma', default=0.5)
    parser.add_argument('-K', '--treedepthsigma', default=0.5)
    #parser.add_argument('-O', '--Output', default="test")
    parser.add_argument('-c', '--cn-num', default=1)
    parser.add_argument('-d', '--del-rate', default=0.5)
    parser.add_argument('-m', '--min-cn-size', default=200000)
    #parser.add_argument('-m', '--min-cn-size', default=10)
    parser.add_argument('-e', '--exp-theta', default=0.000001)
    parser.add_argument('-a', '--amp-p', default=0.5)
    parser.add_argument('-t', '--template-ref', default="hg19.fa")
    #parser.add_argument('-t', '--template-ref', default="chr1.fa")
    parser.add_argument('-o', '--outfile', default="std.out")
    parser.add_argument('-f', '--fa-prefix', default="ref")
    parser.add_argument('-x', '--Lorenz-x', default=0.5)
    parser.add_argument('-y', '--Lorenz-y', default=0.4)
    parser.add_argument('-v', '--coverage', default=0.02)
    #parser.add_argument('-v', '--coverage', default=10)
    parser.add_argument('-l', '--readlen', default=35)
    parser.add_argument('-w', '--window-size', default=200000)
    #parser.add_argument('-w', '--window-size', default=20)
    parser.add_argument('-u', '--acceptance-rate', default=0.5)
    parser.add_argument('-k', '--skip-first-step', default=0)
    parser.add_argument('-R', '--snv_rate', default=0)
    parser.add_argument('-X', '--multi-root', default=4)
    parser.add_argument('-W', '--whole-amp', default=0)
    parser.add_argument('-C', '--whole-amp-rate', default=0.2)
    parser.add_argument('-E', '--whole-amp-num', default=1)
    parser.add_argument('-J', '--amp-num-geo-par', default=1)
    parser.add_argument('-Y', '--leaf-index-range', default="-1")
    #parser.add_argument('-I', '--leaf-ID-range', default="-1")
    parser.add_argument('-L', '--levels', default="-1")
    parser.add_argument('-U', '--bulk-levels', default="NA")
    parser.add_argument('-V', '--cov-bulk', default="30")
    parser.add_argument('-M', '--single-cell-per-clone', default=0)
    
    
    args = parser.parse_args()
    NUM_OF_PROCESSES = int(args.processors)
    skip = int(args.skip_first_step)
    dir = args.directory
    save_prefix = dir + "/" + "from_first_step" 
    wgsim_dir = args.wgsim_dir
    n = int(args.leaf_num)
    Beta = float(args.Beta)
    Alpha = float(args.Alpha)
    Delta = float(args.Delta)
    treeWidth = float(args.treewidth)
    treeDepth = float(args.treedepth)
    treeWidthSigma = float(args.treewidthsigma)
    treeDepthSigma = float(args.treedepthsigma)
    #Output = args.Output
    cn_num = int(args.cn_num)
    del_rate = float(args.del_rate)
    min_cn_size = int(args.min_cn_size)
    exp_theta = float(args.exp_theta)
    amp_p = float(args.amp_p)
    template_ref = args.template_ref
    outfile = dir + "/" + args.outfile
    fa_prefix = dir + "/" + args.fa_prefix
    x0 = float(args.Lorenz_x)
    y0 = float(args.Lorenz_y)
    cov = float(args.coverage)
    l = int(args.readlen)
    window_size = int(args.window_size)
    u = float(args.acceptance_rate)
    #sigma = args.sigma_for_Gaussian
    snv_rate = float(args.snv_rate)
    root_mult = int(args.multi_root)
    whole_amp = int(args.whole_amp)
    whole_amp_rate = float(args.whole_amp_rate)
    whole_amp_num = int(args.whole_amp_num)
    amp_num_geo_par = float(args.amp_num_geo_par)
    # for one cell per clone, specify leaf_index_range instead of leaf_ID_range
    leaf_index_range = args.leaf_index_range
    single_cell_per_clone = (int(args.single_cell_per_clone) == 1) 
    
    index_min = 0
    index_max = 10000000
    if leaf_index_range != "-1":
        index_min, index_max = leaf_index_range.split(".")
        index_min = int(index_min)
        index_max = int(index_max)
    # leaf_ID_range is related with the level_index
    #leaf_ID_range = args.leaf_ID_range
    #leaf_index_min = 0
    #leaf_index_max = 10000000
    #if leaf_ID_range != "-1":
    #    leaf_index_min, leaf_index_max = leaf_ID_range.split(".")
    #    leaf_index_min = int(leaf_index_min)
    #    leaf_index_max = int(leaf_index_max)
    
    # levels for tree inferencer and longitudinal study
    # process level so that the negative ones are turned to positive, the ones over the deepest level will be reported, and all levels will be sorted in a numerical increasing order
    levels = args.levels.split(";")
    bulk_levels = []
    if args.bulk_levels != "NA":
        bulk_levels = args.bulk_levels.split(";")
        cov_bulk = int(args.cov_bulk)
    
    if skip == 0: 
        if not os.path.exists(dir):
            subprocess.check_call("mkdir " + dir, shell=True)
        else:
            print(dir + " exists. Exit. ")
            sys.exit(0)
    
    leaf_chrlen = []
    leaf_index = []
    chr_name_array = []
    
    def check_Ns(file):
        N_line = 0
        total_line = 0
        with open(file, "r") as f:
            for line in f:
                total_line = total_line + 1
                if line.find('N') != -1:
                    N_line = N_line + 1
        f.close()
        if N_line/float(total_line) > 0.2:
            return 0
        return 1
    
    
    
    # Step 1. generate a phylogentic tree that has copy number alterations on branches.
    # Now all CNs are in tree, not generating fa or remember it in the tree nodes.
    if skip == 0:
        print("before generating tree:")
        [tree, tree_elements_arr] = gen_tree(Beta, Alpha, Delta, treeWidth, treeWidthSigma, treeDepth, treeDepthSigma, dir, cn_num, del_rate, min_cn_size, exp_theta, amp_p, template_ref, outfile, fa_prefix, snv_rate, root_mult, whole_amp, whole_amp_rate, whole_amp_num, amp_num_geo_par)
        #[tree, tree_elements_arr] = gen_tree(n, Beta, Alpha, Delta, treeDepth, treeDepthSigma, dir, cn_num, del_rate, min_cn_size, exp_theta, amp_p, template_ref, outfile, fa_prefix, snv_rate, root_mult, whole_amp, whole_amp_rate, whole_amp_num, amp_num_geo_par)
        #[level_chrlens, level_indices, chr_name_array, tree] = gen_tree(n, Beta, Alpha, Delta, treeWidth, treeWidthSigma, treeDepth, treeDepthSigma, dir, cn_num, del_rate, min_cn_size, exp_theta, amp_p, template_ref, outfile, fa_prefix, snv_rate, root_mult, whole_amp, whole_amp_rate, whole_amp_num, amp_num_geo_par)
        #numpy.save(save_prefix + ".tree_elements.npy", tree_elements)
        #numpy.save(save_prefix + ".leaf_chrlen.npy", leaf_chrlen)
        #numpy.save(save_prefix + ".leaf_index.npy", leaf_index)
        #numpy.save(save_prefix + ".level_chrlen.npy", level_chrlens)
        #numpy.save(save_prefix + ".level_index.npy", level_indices)
        #numpy.save(save_prefix + ".chr_name_array.npy", chr_name_array)
        # save the tree for parallele job submission afterwards
        numpy.save(save_prefix + ".tree_elements.npy", tree_elements_arr)
        numpy.save(save_prefix + ".tree.npy", tree)
        #numpy.save(save_prefix + ".ref.npy", ref)
        print("Done with generating the tree. Save to npy. ")
    #print leaf_index
    
    # Step 2. Consider even coverage, use metropolis hasting to sample read count in each bin based on a given point on Lorenz curve. 
    
    # if first step was skipped, read the previous stored file
    if skip == 1:
        print("Skip the first step. Reading ")
        #abs_path = os.getcwd()
        tree_elements_f = save_prefix + ".tree_elements.npy"
        #leaf_chrlen_f = save_prefix + ".leaf_chrlen.npy"
        # level chrlen, instead of having only the chrlen on the leaves, now also record the chrlen on each level
        #level_chrlen_f = save_prefix = ".level_chrlen.npy"
        #leaf_index_f = save_prefix + ".leaf_index.npy"
        # each level leads to the node indices on this level, dict: level (1 for the first tumor cell) -> node indices separated by semi-colon
        #level_index_f = save_prefix + ".level_index.npy"
        #chr_name_array_f = save_prefix + ".chr_name_array.npy"
        tree_f = save_prefix + ".tree.npy"
        #ref_f = save_prefix + ".ref.npy"
        #leaf_chrlen = numpy.load(leaf_chrlen_f)
        # record the chrlen for each level, dict: level (1 for the first tumor cell) -> array with chrlen at this level
        #level_chrlens = numpy.load(level_chrlen_f)
        #leaf_index = numpy.load(leaf_index_f)
        #level_indices = numpy.load(level_index_f)
        #chr_name_array = numpy.load(chr_name_array_f)
        tree_elements_arr = numpy.load(tree_elements_f, allow_pickle=True)
        tree_elements = tree_elements_arr[0]
        tree = numpy.load(tree_f, allow_pickle=True)
        # depth is the level where the leaves are 
        depth = tree_elements.tree_D
        # max_depth is the level whre the leaves are, tree_D is tempoariraly defunct
        max_depth = tree_elements.max_depth
        #depth = tree_elements.get_depth()
        level_chrlens = tree_elements.level_chrlens
        level_indice = tree_elements.level_indice
        chr_name_array = tree_elements.chr_name_array
        [ref, tmp_chr_name, tmp_len_chr] = init_ref(template_ref)
        #ref = numpy.load(ref_f)
        #print(leaf_chrlen_f)
        #print(leaf_index_f)
        #print(chr_name_array_f)
    
        # make it either making a tree, or generating the leaves
        # now alpha and beta are redefined and are for the read coverage uniformity
        [Alpha, Beta] = get_beta_dist(x0, y0)
    
        [Alpha_bulk, Beta_bulk] = get_beta_dist(x0_bulk, y0_bulk)
        print("Bulk sequencing: Alpha = " + str(Alpha_bulk) + ", Beta = " + str(Beta_bulk))
        # start sequencing bulk 
        for bulk_level in bulk_levels:
            index = 0
            original_level = bulk_level
            bulk_level = int(bulk_level)
            if bulk_level < 0:
                bulk_level = depth + bulk_level + 1
            if bulk_level not in level_chrlens:
                sys.exit("Warning: level " + original_level + " is out of range. Please specify a small -K, and use the -G to constrain your specified levels. ") 
            level_chrlen = level_chrlens[bulk_level]
            level_index = level_indice[bulk_level]
            processes = []
            fq_file_names = []
            # the parallel computing is on each node on the same level
            for all_chrlen in level_chrlen:
                make_fa_wABs(level_index[index], tree, ref, chr_name_array, fa_prefix)
                perc = tree[level_index[index]].perc
                ref_files = fa_prefix + str(level_index[index]) + "_*.fa"
                processes.append(mp.Process(target=gen_reads, args=(dir, index, level_index, all_chrlen, fa_prefix, Alpha_bulk, Beta_bulk, x0_bulk, y0_bulk, cov_bulk * perc, l, window_size, u, chr_name_array, -1, original_level, "bulk")))
                index = index + 1
                # when all nodes at this level is sampled, merge them into one big fastq file as this is bulk sequencing
                fq_file_name1 = "level" + original_level + "_node" + level_index[index] + "_allele1.fq" 
                fq_file_names.append(fq_file_name1)
                fq_file_name2 = "level" + original_level + "_node" + level_index[index] + "_allele2.fq" 
                fq_file_names.append(fq_file_name2)
    
            
            bulk_fq = "level" + original_level + "_bulk.fq"
            for p in processes:
                p.start()
            
            for p in processes:
                p.join()
    
            for f in fq_file_names:
                os.system("cat " + f + " >> " + bulk_fq) 
    
            print("Done with generating bulk for level " + original_level)
    
        print("Alpha = %.2f, Beta = %.2f", Alpha, Beta)
        print("Number of processes: " + str(NUM_OF_PROCESSES))
        # assume for each level, the total number of cells remain the same
        print("Number of cells for each level: " + str(n))
    
        # Serial for each level and each node. For each node, make it parallel. 
        for level in levels:
            index = 0
            # process level by level, levels are specified by the user
            # original_level is a string and might be negative. User specified level representation will be in the fq file names.
            original_level = level
            level = int(level)
            if level < 0: 
                # -1 is for leaves, will transform it to leaf's level
                level = max_depth + level + 1
                print("Will sequence level " + str(level))
            if len(level_chrlens[level]) == 0:
                sys.exit("Warning: level " + original_level + " is out of range. Please specify a small -K, and use the -G to constrain your specified levels. ") 
    
            level_chrlen = level_chrlens[level]
            level_index = level_indice[level]
            for all_chrlen in level_chrlen:
                # each node at this level
        #for all_chrlen in leaf_chrlen:
                # each leaf
                # parallelize the whole process on the cells on a clone
                processes = []
            #TODO add a function here to retrieve the CN and apply to reference, and write to a fa file for the two alleles. This has been done.
                #leaf_index_ = leaf_index[index]
                # the index of the node obtained from level_index is w.r.t. the chrlen from level_chrlen
                # this is the node id including interal nodes.
                level_index_ = level_index[index]
                #if leaf_index_range != "-1" and index < index_max and index >= index_min or leaf_ID_range != "-1" and leaf_index_ < leaf_index_max and leaf_index_ >= leaf_index_min or leaf_index_range == "-1" and leaf_ID_range == "-1": 
                #if leaf_index_range != "-1" and index < index_max and index >= index_min or leaf_ID_range != "-1" and level_index_ < leaf_index_max and level_index_ >= leaf_index_min or leaf_index_range == "-1" and leaf_ID_range == "-1": 
                if leaf_index_range != "-1" and index < index_max and index >= index_min or leaf_index_range == "-1": 
                    #make_fa(leaf_index[index], tree, ref, chr_name_array, fa_prefix)
                    make_fa_wABs(level_index[index], tree, ref, chr_name_array, fa_prefix)
                    
                    # since this is a subclone, get the percentage of it
                    #perc = tree[leaf_index[index]].perc
                    perc = tree[level_index[index]].perc
                    cell_num = int(n) * perc
                    # allowing not considering multiple cells in a clone (node in a tree), thus each node has only one cell
                    if single_cell_per_clone:
                        cell_num = 1
                    #total_cell_num += cell_num
                    #TODO need to take care the difference of the added cell number and the total
                    cell_i_last = False
                    for cell_i in range(cell_num):
                        # sequence each cell in this subclone 
                        if cell_i == cell_num - 1:
                            # for removing the fa file
                            cell_i_last = True
                        #processes.append(mp.Process(target=gen_reads, args=(dir, index, leaf_index, all_chrlen, fa_prefix, Alpha, Beta, x0, y0, cov, l, window_size, u, chr_name_array, cell_i)))
                        processes.append(mp.Process(target=gen_reads, args=(dir, index, level_index, all_chrlen, fa_prefix, Alpha, Beta, x0, y0, cov, l, window_size, u, chr_name_array, cell_i, original_level, "sc")))
        
                    # clean the rest
                    for p in processes:
                        p.start()
            
                    for p in processes:
                        p.join()
        
                    if cell_i_last:
                        #print "Done with the whole clone " + str(leaf_index[index]) + ", will remove it. "
                        print("Done with the whole clone " + str(level_index[index]) + ", will remove it. ")
            
                        #ref_files = fa_prefix + str(leaf_index[index]) + "_*.fa"
                        ref_files = fa_prefix + str(level_index[index]) + "_*.fa"
                        args = "rm " + ref_files 
                        os.system(args)
            
                # deal with the next node on this level
                index = index + 1
    
        
        # clean up the temp fa files
        #args = "rm " + fa_prefix + "*_*_*_*_*.fa"
        #popen = subprocess.Popen(args, stdout=subprocess.PIPE, shell=True)
        #popen.wait()

if __name__ == "__main__":
    main()



#!/usr/bin/python$
import numpy
import copy
from gen_tree import gen_tree 
from Gen_Ref_Fa import getlen_ref
from gen_tree import get_cn_from_corres
import CN
import argparse
import sys
import re

class mynode:
    def __init__(self, ID, parent=None):
        self.ID = ID
        # ID
        self.parent = parent
        self.children = []

# print all snvs if no node is specified
def print_snvs(tree, cell):
    print("\t".join(["chr", "pos", "ori_nuc", "new_nuc", "allele", "cell"]))
    cells = []
    if cell == "NA":
        for i in range(len(tree)):
            cells.append(i)
    else:
        array = re.split(r';', cell)
        for i in array:
            cells.append(int(i))
    
    for i in cells:
        snv = tree[i].snvs
        for s in snv:
            print("\t".join([str(s.chr), str(s.ref_pos), s.ori_nuc, s.new_nuc, str(s.ale), str(i)]))
    

# a new function added in Feb. 2020
# output a matrix, in which the entry of (i, j) represents the number of events between these two profiles (here assume i and j are both leaves)
def get_pairwise_diff_matrix(tree):
    leaves = get_leafid_array(tree)
    for i in range(len(leaves)):
        for j in range(len(leaves)):
            x = get_pairwise_diff(tree, leaves[i], leaves[j])
            print(str(x) + "\t", end="")
        print("")
    

# a new function added in Feb. 2020
# for two leaf cells, output their difference (different event #)
def get_pairwise_diff(tree, i, j):
    p = LCA(tree, i, j)
    #print "LCA of " + str(i) + " and " + str(j) + " is " + str(p)
    e_num_i = get_event_num_path(tree, i, p)
    e_num_j = get_event_num_path(tree, j, p)
    return e_num_i + e_num_j
    
# augmentary function
# given two nodes, get the least common ancestor
def LCA(tree, i, j): 
    P_i = get_path(tree, i)
    P_j = get_path(tree, j)
    for x in P_i:
        for y in P_j:
            if x == y:
                return x
     
# augmentary function
# given the tree and a node, find the path bottom up till it reaches the root
def get_path(tree, i):
    p = tree[i].parent.id
    P = [i]
    while p != -1:
        P.append(p)
        #print p
        p = tree[p].parent.id
    return P 
 
# augmentary funciton
# given a tree, two nodes, one is the ancestor of the other, calculate the total event number from the ancestor to the node
def get_event_num_path(tree, i, p):
    cn_array_num = 0
    while i != p:
        cn_array_num += len(tree[i].cn)
        i = tree[i].parent.id
    return cn_array_num

# a new function added in Jan. 2020
# for each daughter cell compared with a parent cell, output its new CNAs
def get_event_num(tree):
    for t in tree:
        ID = t.id
        cn_array = t.cn
        print(str(ID) + "\t" + str(len(cn_array)) + "\t" + str(t.if_leaf))


def get_leaf(tree):
    for t in tree:
        if t.if_leaf:
            print(str(t.id))

def get_leafid_array(tree):
    a = []
    for t in tree:
        if t.if_leaf:
            a.append(t.id)
    return a

def retrieve_new_CNAs(tree):
    s = {}
    S = {}
    for t in tree:
        ID = t.id
        s[ID] = str(ID)
        S[ID] = str(ID) 
    # now count
    for t in tree:
        ID = t.id
        # s contains only new CNAs to this node
        s[ID] = {}
        # S contains all
        S[ID] = {}
        summary = t.cn_summary
        p = t.parentID        
        for chr in summary:
            for pos in summary[chr]:
                chr_ = chr + 1
                if chr_ == 23:
                    chr_ = "X"
                if chr_ == 24:
                    chr_ = "Y"
                str_ = str(chr_) + ":" + str(pos) + "_" + str(summary[chr][pos])
                if p == -1:
                    s[ID][str_] = 1
                elif str_ not in S[p]:
                    s[ID][str_] = 1
                S[ID][str_] = 1
    return s

# given a segcopy file with all nodes, a tree with the parent-children relationship, output the new CNAs of a child compared to the parent, including internal nodes
def retrieve_new_overlappingCNAs(segcopy_f, Tree):
    f = open(segcopy_f, "r")
    line = f.readline().rstrip("\n")
    names = []
    # from leaf id to the column index
    names_h = {}
    # h is from the position to the leaf index (0-based) to CN
    h = {}
    first = True
    while(line != ""):
        array = re.split(r'\s+', line)
        if first:
            names = array[3:]
            for i in range(len(names)):
                if names[i] == "":
                    break
                names_h[int(names[i][4:])] = i
            first = False
        else:
            cnas = array[3:]
            loc = str(array[0]) + ":" + str(array[1]) + "-" + str(array[2])
            h[loc] = {}
            #locs.append(loc)
            for i in range(len(cnas)):
                if cnas[i] == "":
                    break
                h[loc][i] = int(cnas[i])
        line = f.readline().rstrip("\n")
    f.close()
    # now retrieve the new CNAs comparing child's with parent's CNA profile
    # new_cna_h is from the leaf id to chromosome to positions to CNA (+: increase, -: decrease) with the absolute copy number change
    new_cna_h = {}
    for i in range(len(Tree)):
        p = Tree[i].parentID
        new_cna_h = {}
        i_col_id = names_h[i]
        if p == -1:
            for k in list(h.keys()):
                chr, pos = re.split(r':', k)
                s, e = re.split(r'-', pos)
                s = int(s)
                if chr not in list(new_cna_h.keys()):
                    new_cna_h[chr] = {}
                if h[k][i_col_id] != 2:
                    new_cna_h[chr][s] = e + ";" + str(h[k][i_col_id] - 2)
        else:
            p_col_id = names_h[p]
            for k in list(h.keys()):
                diff = h[k][i_col_id] - h[k][p_col_id] 
                if diff != 0:
                    chr, pos = re.split(r':', k)
                    s, e = re.split(r'-', pos)
                    s = int(s)
                    if chr not in list(new_cna_h.keys()):
                        new_cna_h[chr] = {}
                    new_cna_h[chr][s] = e + ";" + str(diff)
        combined = combine_cnas(new_cna_h)
        for l in combined:
            a = copy.deepcopy(l)
            a.append(str(i))
            a.append(str(p))
            print("\t".join(a))

# connect the CNA bins together, return an array with chr, start, end
def combine_cnas(h):
    h_ret = []
    for chr in list(h.keys()):
        start = "NA"
        end = "NA"
        prev_e = "NA"
        prev_cn = "NA"
        interval = 0
        for ss in sorted(list(h[chr].keys())):
            s = str(ss)
            e, cn = re.split(r';', h[chr][ss])
            if start == "NA":
                start = s
                end = e
                prev_e = e
                prev_cn = cn
                interval = 0
            elif int(s) == int(prev_e) + 1 and prev_cn == cn:
                interval += 1
                end = e
                prev_e = e
            else:
                # temporarily not look at the actual cn 
                if interval > 1:
                    h_ret.append([chr, start, end, prev_cn])
                start = s
                end = e
                prev_e = e
                prev_cn = cn
                interval = 0

        if interval > 1:
            h_ret.append([chr, start, end, cn])
    return h_ret
    
            
# print the tree in newick format, level is to control when to stop to dig into the tree (in which case only leaves in that cluster and the head node will be printed)
def convert2newick(Tree, str_, i):
    # stop
    if i >= len(Tree):
        return str_
    node = Tree[i]
    parent = node.parent
    children = node.children
    # get rid of the node itself
    c_arr = []
    for c in children:
        if c != str(i):
            c_arr.append(c)
    children = c_arr

    if len(children) >= 1:
        #print children
        # prepare for where to replace
        splitted = re.split(r'(\d+)', str_)
        i_ = -1
        for i_ in range(len(splitted)):
            if str(splitted[i_]) == str(i):
                break
        # now make a new string to replace
        new_substr = "(" + "," . join(children) + ")"
        #print str(splitted[i_]) + " is to be replaced by " + new_substr
        # concatenate
        splitted[i_] = new_substr + splitted[i_]
        str_ = "".join(splitted)
        str_ = convert2newick(Tree, str_, i + 1)
    else:
        str_ = convert2newick(Tree, str_, i + 1)
    return str_
        
        
def print_newick(tree):

    # a new tree with a simplified structure
    Tree = []
    
    # step 1, read the npy file and put them in MyNode
    for i in range(len(tree)):
        node = tree[i]
        parent = node.parentID
        ID = i
        Tree.append(mynode(ID, parent))
        # update the children of its parent
        Tree[parent].children.append(str(ID))
    
    str_ = convert2newick(Tree, "(0)", 0)
    #print str[1:-1]
    return str_

def print_tree(tree):
    leaf_only = True
    dec = get_descendants(tree, leaf_only, False, "NA") 
    s = retrieve_new_CNAs(tree)
    children = get_children(tree)
    for t in tree:
        ID = t.id
        #if ID == 0:
        #    continue
        #else:
        decs = dec[ID]
        num = len(decs)
        print(str(ID) + "\t" + ";".join(children[ID]) + "\t" + str(num) + "\t" + ";".join(list(s[ID].keys())) + "\t" + ";".join(decs))

def get_summary(tree, select_leaf):
    for t in tree:
        ID = t.id
        if select_leaf:
            if not t.if_leaf:
                continue
        summary = t.cn_summary
        for chr in summary:
            for pos in summary[chr]:
                pos_s, pos_e = pos.split(".")
                cn = summary[chr][pos]
                chr_ = chr + 1
                chr_ = "chr" + str(chr_)
                if chr_ == "chr23":
                    chr_ = "chrX"
                if chr_ == "chr24":
                    chr_ = "chrY"
                print("\t".join([chr_, str(pos_s), str(pos_e), str(cn), str(ID)])) 

def print_large_clusters(tree, leaf_only, size):
    d = get_descendants(tree, leaf_only, True, size)
    for i in d:
        if len(d[i]) != 0:
            y = [str(x) for x in d[i]]
            print(" ".join(y))

def retrieve_new_CNAs(tree):
    for i in range(len(tree)):
        t = tree[i]
        tCNs = t.true_CNs
        for tCN in tCNs: 
            for k in list(tCN.keys()):
                chr, interval = k.split(":")
                s, e = interval.split(".")
                print("\t".join([chr, s, e, str(tCN[k]), str(i), str(t.parentID)])) 
    return
        
# for each node in the tree, find the descendants of it
# in a reverse order, traverse the tree. For every node, add the descendants of its daughter cells, and its daughter cells into its descendant list. 
# if leaf_only is on, output only leaf as the descendants; 
# if cut_by_size is on, then remove the descendants if the parent's descendants do not fall into the size range
def get_descendants(tree, leaf_only, cut_by_size, size):
    d = {}
    # intiialize the dictionary
    for t in tree[::-1]:
        ID = t.id
        if leaf_only:
            if t.if_leaf:
                d[ID] = [str(ID)]
            else:
                d[ID] = []
        else:
            d[ID] = [str(ID)]
    # now count
    for t in tree[::-1]:
        ID = t.id
        p = t.parentID
        if p != -1:
            for i in d[ID]:
                d[p].append(i) 

    # cut by size
    if cut_by_size:
        min_, max_ = size.split(",")
        for t in tree:
            if len(d[t.id]) < int(min_) or len(d[t.id]) > int(max_):
                d[t.id] = []
                
    return d

# return a dictionary with the keys the ids, the values an array of children IDs 
def get_children(tree):
    d = {}
    for t in tree:
        ID = t.id
        d[ID] = []
    for t in tree:
        ID = t.id
        p = t.parentID
        if p != -1:
            d[p].append(str(ID))
    return d

# in case the tree does not have the cn summary 
def make_summary_func(tree, ref):
    tmp_name_array, chr_sz = getlen_ref(ref)
    for i in tree:
        i.cn_detail, i.cn_summary = get_cn_from_corres(i.corres, chr_sz)
        ID = i.id
        for chr in i.cn_summary:
            for pos in i.cn_summary[chr]:
                pos_s, pos_e = pos.split(".")
                cn = i.cn_summary[chr][pos]
                chr_ = chr + 1
                chr_ = "chr" + str(chr_)
                if chr_ == "chr23":
                    chr_ = "chrX"
                if chr_ == "chr24":
                    chr_ = "chrY"
                print("\t".join([chr_, str(pos_s), str(pos_e), str(cn), str(ID)])) 


parser = argparse.ArgumentParser(description='Read a tree and output specific items of it. ')
parser.add_argument('-l', '--leaf', action='store_true')  
parser.add_argument('-s', '--summary', action='store_true')  
parser.add_argument('-L', '--select-leaf', action='store_true')  
parser.add_argument('-m', '--make-summary', action='store_true')
parser.add_argument('-r', '--ref', default="")
parser.add_argument('-f', '--file', default="") 
parser.add_argument('-e', '--event', action='store_true')
parser.add_argument('-d', '--pairdist', action='store_true')
parser.add_argument('-P', '--printtree', action='store_true')
parser.add_argument('-N', '--printnewick', action='store_true')
parser.add_argument('-C', '--printlargecluster', action='store_true')
parser.add_argument('-S', '--largeclustersize', default="8,10")
parser.add_argument('-O', '--retrieveoverlapping', action='store_true')
parser.add_argument('-F', '--segcopyf', default="NA")
parser.add_argument('-o', '--retrievealloverlaps', action='store_true')
parser.add_argument('-n', '--printsnvs', action='store_true')
parser.add_argument('-x', '--printsnvsforcell', default="NA")


args = parser.parse_args()
if_leaf = args.leaf
event_num = args.event
pairdist = args.pairdist
if_summary = args.summary
select_leaf = args.select_leaf
make_summary = args.make_summary
printtree = args.printtree
ref_f = args.ref
npy_f = args.file
printnewick = args.printnewick
printlargecluster = args.printlargecluster
cluster_size = args.largeclustersize
retrieveoverlapping = args.retrieveoverlapping
segcopy_f = args.segcopyf
retrievealloverlaps = args.retrievealloverlaps
printsnvs = args.printsnvs
printsnvsforcell = args.printsnvsforcell

# main starts here
if npy_f == "": 
    print("""
    Given a tree in npy format, output its leaf index or the CNV summary. 
    esage: python read_tree.py -l -s -e -d -o -f [tree.npy]
        -l  (--leaf)    Print leaf index, one per line. (default: off)
        -s  (--summary) Print CNV summary, one per line (chr, start, end, CN). (default: off)
        -L  (--selectleaf)  select to print leaf, in conjunction with if_summary. (default: off)
        -m  (--make-summary)    When the tree does not have CNV summary info but the correspondence, make it and print the CNV info, like -s. Need the -r info to retrieve the info back. (default: off)
        -r  (--ref)     Reference file in .fa. Necessary only when -m is turned on. 
        -f  (--file)    The npy file storing the tree. (mandatory)
        -e  (--event)   Get the number of events (col 2) for each node (col 1) and indicate whether it is a leaf node or not (col 3). 
        -d  (--pairdist)    Get the pairwise distance between all leaf cells. Output the matrix to the stdout. (default: off) 
        -P  (--printtree)   Print the tree with each line representing a new CNA, with the columns node, the new CNA to this node and its descendants including itself.
        -N  (--printnewick) Print the newick formatted tree.
        -C  (--printlargecluster)   Print large clusters.
        -S  (--largeclustersize)   Size range separated by comma.
        -O  (--retrieveoverlapping) Retrieve new overlapping CNAs for each cell (including internal nodes). 
        -F  (--segcopyf)    Segcopy file for retrieving new overlapping CNA. 
        -o  (--retrievealloverlaps)  Retrieve new overlapping CNAs, even for those occurring on the same edge from true_CNs in the tree. Output the new CNAs for each node compared with its parent node. 
        -n  (--printsnvs)   Print SNVs for all cells.
        -x  (--printsnvsforcell)    Print SNVs for a particular set of nodes separated by colon specified here. 
        """)
    sys.exit(0)


tree = numpy.load(npy_f, allow_pickle=True)
if if_leaf:
    get_leaf(tree)

if if_summary:
    get_summary(tree, select_leaf) 

if make_summary:
    make_summary_func(tree, ref_f)

if event_num:
    get_event_num(tree)

if pairdist:
    get_pairwise_diff_matrix(tree)
    
if printtree:
    print_tree(tree)

if printnewick:
    print(print_newick(tree))

if printlargecluster:
    print_large_clusters(tree, True, cluster_size)

if retrieveoverlapping:
    retrieve_new_overlappingCNAs(segcopy_f, tree)

if retrievealloverlaps:
    retrieve_new_CNAs(tree)

if printsnvs:
    print_snvs(tree, "NA")

if printsnvsforcell != "NA":
    print_snvs(tree, printsnvsforcell) 
